"""
The shared base classes for chat group interactions.

Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE
"""

from __future__ import annotations

import abc
import asyncio
from collections.abc import Coroutine, Sequence
from datetime import datetime
from random import randint
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from typing_extensions import Self, TypeAlias

from . import utils
from .abc import Channel, Message, SteamID
from .enums import ChatMemberRank, Type
from .errors import WSException
from .game import StatefulGame
from .iterators import ChatHistoryIterator
from .protobufs import MsgProto, chat
from .reaction import Emoticon, MessageReaction, Sticker
from .role import Role
from .user import User, WrapsUser
from .utils import DateTime

if TYPE_CHECKING:
    from .channel import ClanChannel, GroupChannel
    from .clan import Clan
    from .group import Group
    from .image import Image
    from .message import Authors, ClanMessage, GroupMessage
    from .state import ConnectionState
    from .types.id import ID32, ID64, ChannelID, ChatGroupID


ChatT = TypeVar("ChatT", bound="Chat[Any]", covariant=True)
MemberT = TypeVar("MemberT", bound="Member", covariant=True)


class Member(WrapsUser):
    """Represents a member of a chat group."""

    __slots__ = (
        "clan",
        "group",
        "rank",
        "kick_expires_at",
        "_role_ids",
    )
    clan: Clan | None
    group: Group | None

    def __init__(
        self, state: ConnectionState, chat_group: ChatGroup[Any, Any], user: User, member: chat.Member
    ) -> None:
        super().__init__(state, user)
        self.clan = None
        self.group = None
        self._update(member)

    def _update(self, member: chat.Member) -> None:
        self.rank = ChatMemberRank.try_value(member.rank)
        self._role_ids = tuple(member.role_ids)
        self.kick_expires_at = DateTime.from_timestamp(member.time_kick_expire)

    @property
    def roles(self) -> list[Role]:
        """The member's roles."""
        chat_group = self.group or self.clan
        assert chat_group is not None
        return [chat_group._roles[role_id] for role_id in self._role_ids]

    # async def add_role(self, role: Role):
    #     """Add a role to the member."""

    # async def ban(self, *, reason: str | None = None) -> None:
    #     """Bans the member from the chat group."""
    #     group = self.group or self.clan
    #     assert group is not None
    #     await group.ban(self, reason=reason)


class ChatMessage(Message):
    channel: GroupChannel | ClanChannel
    mentions: chat.Mentions
    author: Member | SteamID

    def __init__(self, proto: chat.IncomingChatMessageNotification, channel: Any, author: Authors):
        super().__init__(channel, proto)
        self.author = author
        self.created_at = DateTime.from_timestamp(proto.timestamp)

    async def fetch_reaction(self, emoticon: Emoticon | Sticker) -> list[MessageReaction]:
        """Fetches the reactions to this message with a given emoticon."""
        chat_group = self.clan or self.group
        assert chat_group is not None
        is_emoticon = isinstance(emoticon, Emoticon)
        reactors = await self._state.fetch_message_reactors(
            *self.channel._location,
            server_timestamp=int(self.created_at.timestamp()),
            ordinal=self.ordinal,
            reaction_name=str(emoticon) if is_emoticon else emoticon.name,
            reaction_type=1 if is_emoticon else 2,
        )

        reaction = utils.find(
            lambda r: (r.emoticon or r.sticker).name == emoticon.name, self.partial_reactions  # type: ignore
        )
        assert reaction
        return [
            MessageReaction(
                self._state,
                self,
                reaction.emoticon,
                reaction.sticker,
                user=chat_group._members.get(reactor) or SteamID(reactor),
            )
            for reactor in reactors
        ]

    async def fetch_reactions(self) -> list[MessageReaction]:
        """Fetches all this message's reactions.

        Warning
        -------
        This does nothing for messages that aren't fetched by :meth:`GroupChannel.history`/:meth:`ClanChannel.history`.
        """
        return [
            r
            for rs in await asyncio.gather(
                *(
                    self.fetch_reaction(reaction.emoticon or reaction.sticker)  # type: ignore
                    for reaction in self.partial_reactions
                )
            )
            for r in rs
        ]

    async def add_emoticon(self, emoticon: Emoticon) -> None:
        await self._state.react_to_chat_message(
            *self.channel._location,
            int(self.created_at.timestamp()),
            self.ordinal,
            str(emoticon),
            reaction_type=1,
            is_add=True,
        )
        self._state.dispatch(
            "reaction_add",
            MessageReaction(self._state, self, emoticon, None, self._state.user, DateTime.now(), self.ordinal),
        )

    async def remove_emoticon(self, emoticon: Emoticon) -> None:
        await self._state.react_to_chat_message(
            *self.channel._location,
            int(self.created_at.timestamp()),
            self.ordinal,
            str(emoticon),
            reaction_type=1,
            is_add=False,
        )
        self._state.dispatch(
            "reaction_remove",
            MessageReaction(self._state, self, emoticon, None, self._state.user, DateTime.now(), self.ordinal),
        )

    async def add_sticker(self, sticker: Sticker) -> None:
        await self._state.react_to_chat_message(
            *self.channel._location,
            int(self.created_at.timestamp()),
            self.ordinal,
            sticker.name,
            reaction_type=2,
            is_add=True,
        )
        self._state.dispatch(
            "reaction_add",
            MessageReaction(self._state, self, None, sticker, self._state.user, DateTime.now(), self.ordinal),
        )

    async def remove_sticker(self, sticker: Sticker) -> None:
        await self._state.react_to_chat_message(
            *self.channel._location,
            int(self.created_at.timestamp()),
            self.ordinal,
            sticker.name,
            reaction_type=2,
            is_add=False,
        )
        self._state.dispatch(
            "reaction_remove",
            MessageReaction(self._state, self, None, sticker, self._state.client.user, DateTime.now(), self.ordinal),
        )


ChatMessageT = TypeVar("ChatMessageT", bound="GroupMessage | ClanMessage", covariant=True)
GroupChannelProtos: TypeAlias = "chat.IncomingChatMessageNotification | chat.State | chat.ChatRoomState"


class Chat(Channel[ChatMessageT]):
    __slots__ = ("id", "name", "joined_at", "position", "last_message")

    def __init__(
        self, state: ConnectionState, group: ChatGroup[Any, Self], proto: GroupChannelProtos
    ):  # group is purposely unused
        super().__init__(state)
        self.id = int(proto.chat_id)
        self.name: str | None = None
        self.joined_at: datetime | None = None
        self.position: int | None = None
        self.last_message: ChatMessageT | None = None
        self._update(proto)

    def _update(self, proto: GroupChannelProtos) -> None:
        if isinstance(proto, (chat.State, chat.IncomingChatMessageNotification)):
            first, _, second = proto.chat_name.partition(" | ")
            self.name = second or first

        self.joined_at = (
            DateTime.from_timestamp(proto.time_joined) if isinstance(proto, chat.ChatRoomState) else self.joined_at
        )
        self.position = proto.sort_order if isinstance(proto, chat.State) else self.position
        (message_cls,) = self._type_args
        if isinstance(proto, chat.IncomingChatMessageNotification):
            steam_id = SteamID(proto.steamid_sender, type=Type.Individual)
            self.last_message = message_cls(proto, self, self._state.get_user(steam_id.id) or steam_id)  # type: ignore

    def __repr__(self) -> str:
        attrs = ("name", "id", "group" if self.group is not None else "clan", "position")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __eq__(self, other: object) -> bool:
        return self._location == other._location if isinstance(other, Chat) else NotImplemented

    @property
    def _location(self) -> tuple[ChatGroupID, ChannelID]:
        chat = self.clan or self.group
        assert chat is not None
        chat_id = chat._id
        assert chat_id is not None
        return chat_id, self.id

    @utils.classproperty
    def _type_args(cls: type[Chat]) -> tuple[type[ChatMessageT]]:  # type: ignore
        return cls.__orig_bases__[0].__args__  # type: ignore

    def _message_func(self, content: str) -> Coroutine[Any, Any, ChatMessageT]:
        return self._state.send_chat_message(*self._location, content)  # type: ignore

    def _image_func(self, image: Image) -> Coroutine[Any, Any, None]:
        return self._state.http.send_chat_image(*self._location, image)

    def history(
        self,
        *,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> ChatHistoryIterator[ChatMessageT, Self]:
        return ChatHistoryIterator(state=self._state, channel=self, limit=limit, before=before, after=after)

    # async def edit(self, *, name: str) -> None:
    #     await self._state.edit_channel(*self._location, name)


# class SteamIDMember(SteamID):  # TODO?
#     ...


class ChatGroup(SteamID, Generic[MemberT, ChatT]):
    """Base class for :class:`steam.Clan` and :class:`steam.Group`."""

    __slots__ = (
        "name",
        "_id",
        "game",
        "tagline",
        "avatar_url",
        "active_member_count",
        "chunked",
        "_state",
        "_members",
        "_partial_members",
        "_owner_id",
        "_top_members",
        "_channels",
        "_default_channel_id",
        "_roles",
        "_default_role_id",
    )
    name: str
    _id: ChatGroupID
    game: StatefulGame | None
    tagline: str
    avatar_url: str
    active_member_count: int
    chunked: bool
    _state: ConnectionState

    _members: dict[ID32, MemberT]
    _owner_id: ID32
    _top_members: list[ID32]
    _partial_members: dict[ID32, chat.Member]  # deleted after _members is populated (if chunked)

    _channels: dict[int, ChatT]
    _default_channel_id: int

    _roles: dict[int, Role]
    _default_role_id: int

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.chunked = False
        self.game: StatefulGame | None = None
        self._members = {}
        self._partial_members = {}
        self._channels: dict[int, ChatT] = {}
        self._roles: dict[int, Role] = {}

    @classmethod
    async def _from_proto(
        cls,
        state: ConnectionState,
        proto: chat.GetChatRoomGroupSummaryResponse,
        *,
        id: ID32 | ChatGroupID | None = None,
        maybe_chunk: bool = True,
    ) -> Self:
        self = cls(state, id if id is not None else proto.chat_group_id)
        self.name = proto.chat_group_name
        self._id = proto.chat_group_id
        self.active_member_count = proto.active_member_count
        self._owner_id = proto.accountid_owner
        self._top_members = proto.top_members
        self.tagline = proto.chat_group_tagline
        self.game = StatefulGame(state, id=proto.appid) if proto.appid else self.game

        self._default_role_id = proto.default_role_id
        self._update_channels(proto.chat_rooms, default_channel_id=proto.default_chat_id)

        if maybe_chunk:
            try:
                group_state = await self._state.set_chat_group_active(proto.chat_group_id)
            except WSException:
                pass
            else:
                self._partial_members = {member.accountid: member for member in group_state.members}
                self._roles = {
                    role.role_id: Role(self._state, self, role, permissions)  # type: ignore
                    for role in group_state.header_state.roles
                    for permissions in group_state.header_state.role_actions
                    if permissions.role_id == role.role_id
                }

            if self._state.auto_chunk_chat_groups:
                await self.chunk()

        return self

    @utils.classproperty
    def _type_args(cls: type[ChatGroup]) -> tuple[type[MemberT], type[ChatT]]:  # type: ignore
        return cls.__orig_bases__[0].__args__  # type: ignore

    async def _add_member(self, member: chat.Member) -> MemberT:
        member_cls, _ = self._type_args
        user = await self._state._maybe_user(member.accountid)
        assert isinstance(user, User)
        new_member = member_cls(self._state, self, user, member)
        if self.chunked:
            self._members[member.accountid] = new_member
        else:
            self._partial_members[member.accountid] = member
        return new_member

    def _remove_member(self, member: chat.Member) -> MemberT | None:
        if self.chunked:
            return self._members.pop(member.accountid, None)
        else:
            self._partial_members.pop(member.accountid, None)

    def _update_channels(
        self, channels: list[chat.State] | list[chat.ChatRoomState], *, default_channel_id: int | None = None
    ) -> None:
        _, channel_cls = self._type_args
        for channel in channels:
            try:
                new_channel = self._channels[channel.chat_id]
            except KeyError:
                new_channel = channel_cls(self._state, self, channel)
                self._channels[new_channel.id] = new_channel
            else:
                new_channel._update(channel)

        if default_channel_id is not None:
            self._default_channel_id = default_channel_id

    def _update_header_state(self, proto: chat.GroupHeaderState) -> None:
        self.name = proto.chat_name
        self._owner_id = proto.accountid_owner
        self.game = StatefulGame(self._state, id=proto.appid)
        self.tagline = proto.tagline
        self.avatar_url = utils._get_avatar_url(proto.avatar_sha)
        self._default_role_id = proto.default_role_id or self._default_role_id
        for role in proto.roles:
            for role_action in proto.role_actions:
                if role.role_id == role_action.role_id:
                    self._roles[role.role_id] = Role(self._state, self, role, role_action)  # type: ignore

    def __repr__(self) -> str:
        attrs = ("name", "id", "universe", "instance")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __str__(self) -> str:
        return self.name

    @property
    def members(self) -> Sequence[MemberT]:
        """A list of the chat group's members."""
        return list(self._members.values())

    def get_member(self, id64: ID64) -> MemberT | None:
        """Get a member of the chat group by id.

        Parameters
        ----------
        id64
            The 64 bit ID of the member to get.
        """
        return self._members.get(id64 & 0xFFFFFFFF)  # TODO consider just passing ID32

    @property
    def owner(self):
        return self._members.get(self._owner_id) or SteamID(self._owner_id)

    @property
    def top_members(self) -> Sequence[MemberT | SteamID]:
        """A list of the chat group's top members."""
        return [(self._members.get(id) or SteamID(id)) for id in self._top_members]

    @property
    def me(self) -> MemberT:
        """The client user's account in this chat group."""
        return self._members[self._state.user.id]

    @property
    def channels(self) -> Sequence[ChatT]:
        """A list of the chat group's channels."""
        return list(self._channels.values())

    def get_channel(self, id: int) -> ChatT | None:
        """Get a channel from cache.

        Parameters
        ----------
        id
            The ID of the channel.
        """
        return self._channels.get(id)

    @property
    def default_channel(self) -> ChatT:
        """The group's default channel."""
        return self._channels[self._default_channel_id]

    @property
    def roles(self) -> Sequence[Role]:
        """A list of the group's roles."""
        return list(self._roles.values())

    def get_role(self, id: int) -> Role | None:
        """Get a role from cache.

        Parameters
        ----------
        id
            The ID of the role.
        """
        return self._roles.get(id)

    @property
    def default_role(self) -> Role:
        """The group's default role."""
        return self._roles[self._default_role_id]

    @abc.abstractmethod
    async def chunk(self) -> Sequence[MemberT]:
        """Get a list of all members in the group."""
        self.chunked = True
        del self._partial_members
        return self.members

    async def search(self, name: str) -> list[MemberT]:
        """Search for members in the chat group.

        Parameters
        ----------
        name
            The name of the member to search for.
        """
        msg: MsgProto[chat.SearchMembersResponse] = await self._state.ws.send_um_and_wait(
            "ChatRoom.SearchMembers",
            chat_group_id=self._id,
            search_text=name,
        )

        if self.chunked:
            return [self._members[user.accountid] for user in msg.body.matching_members]

        member_cls, _ = self._type_args
        users = [
            User(self._state, self._state.patch_user_from_ws({}, user.persona)) for user in msg.body.matching_members
        ]
        return [member_cls(self._state, self, user, self._partial_members[user.id]) for user in users]

    async def invite(self, user: User) -> None:
        """Invites a :class:`~steam.User` to the chat group.

        Parameters
        -----------
        user
            The user to invite to the chat group.
        """
        await self._state.invite_user_to_chat_group(user.id64, self._id)

    async def join(self, *, invite_code: str | None = None) -> None:
        """Joins the chat group.

        Parameters
        ----------
        invite_code
            The invite code to use to join the chat group.
        """
        await self._state.join_chat_group(self._id, invite_code)
        if self._state.auto_chunk_chat_groups:
            await self.chunk()

    async def leave(self) -> None:
        """Leaves the chat group."""
        await self._state.leave_chat_group(self._id)

    # TODO
    # async def edit(self, *, name: str | None = None, tagline: str | None = None, avatar: bytes | SupportsRead) -> None:
    #     """Edits the chat group."""
    #     await self._state.edit_chat_group(self._id, name, tagline, avatar)
    #
    # async def create_role(self) -> Role:  # chat.CreateRoleRequest
    #     ...
    #
    # async def create_channel(self) -> ChannelT:  # chat.CreateChatRoomRequest
    #     ...
    #
    # async def create_invite(self) -> InviteLink:  # chat.CreateInviteLinkRequest
    #     ...
    #
    # async def fetch_invite(self, code: str) -> InviteLink:  # chat.GetChatInviteLinkResponse
    #     ...
    #
    # async def invites(self) -> list[InviteLink]:  # chat.GetChatInvitesResponse
    #     ...
    #
    # add Member.x equivalent for these 3
    # async def mute(self, member: SteamID) -> None:  # chat.MuteUserRequest
    #     ...
    #
    # async def kick(self, member: SteamID) -> None:  # chat.KickUserRequest
    #     ...
    #
    # async def ban(self, member: SteamID) -> None:  # chat.SetUserBanStateRequest
    #     ...
    #
    # async def bans(self) -> list[Ban]:  # chat.GetBanListRequest
    #     ...
    #
    # async def unban(self, user: SteamID) -> None:  # chat.SetUserBanStateRequest
    #     ...



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import base64
import dataclasses
import hmac
import struct
from hashlib import sha1
from time import time
from typing import TYPE_CHECKING, Any

from ._const import URL
from .errors import ConfirmationError
from .utils import Intable

if TYPE_CHECKING:
    from .state import ConnectionState


__all__ = (
    "generate_one_time_code",
    "generate_confirmation_code",
    "generate_device_id",
    "Confirmation",
)


def generate_one_time_code(shared_secret: str, timestamp: int | None = None) -> str:
    """Generate a Steam Guard code for signing in or at a specific time.

    Parameters
    -----------
    shared_secret
        Shared secret from steam guard.
    timestamp
        The unix timestamp to generate the key for.
    """
    timestamp = timestamp or int(time())
    time_buffer = struct.pack(">Q", timestamp // 30)  # pack as Big endian, uint64
    time_hmac = hmac.new(base64.b64decode(shared_secret), time_buffer, digestmod=sha1).digest()
    begin = time_hmac[19] & 0xF

    full_code = struct.unpack(">I", time_hmac[begin : begin + 4])[0] & 0x7FFFFFFF  # unpack as Big endian uint32

    chars = "23456789BCDFGHJKMNPQRTVWXY"
    code = []
    for _ in range(5):
        full_code, i = divmod(full_code, len(chars))
        code.append(chars[i])
    return "".join(code)


def generate_confirmation_code(identity_secret: str, tag: str, timestamp: int | None = None) -> str:
    """Generate a trade confirmation code.

    Parameters
    -----------
    identity_secret
        Identity secret from steam guard.
    tag
        Tag to encode to.
    timestamp
        The time to generate the key for.
    """
    timestamp = timestamp or int(time())
    buffer = struct.pack(">Q", timestamp) + tag.encode("ascii")
    return base64.b64encode(hmac.new(base64.b64decode(identity_secret), buffer, digestmod=sha1).digest()).decode()


def generate_device_id(user_id64: Intable) -> str:
    """Generate the device id for a user's 64-bit ID.

    Parameters
    -----------
    user_id64
        The 64 bit steam id to generate the device id for.
    """
    # it works, however it's different that one generated from mobile app

    hexed_steam_id = sha1(str(int(user_id64)).encode("ascii")).hexdigest()
    partial_id = [
        hexed_steam_id[:8],
        hexed_steam_id[8:12],
        hexed_steam_id[12:16],
        hexed_steam_id[16:20],
        hexed_steam_id[20:32],
    ]
    return f'android:{"-".join(partial_id)}'


@dataclasses.dataclass
class Confirmation:
    __slots__ = (
        "_state",
        "id",
        "data_conf_id",
        "data_key",
        "trade_id",
    )

    _state: ConnectionState
    id: str
    data_conf_id: int
    data_key: str
    trade_id: int  # this isn't really always the trade ID, but for our purposes this is fine

    def __repr__(self) -> str:
        return f"<Confirmation id={self.id!r} trade_id={self.trade_id}>"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Confirmation) and self.trade_id == other.trade_id and self.id == other.id

    @property
    def tag(self) -> str:
        return f"details{self.data_conf_id}"

    async def _confirm_params(self, tag: str) -> dict[str, str | int]:
        code, timestamp = await self._state._generate_confirmation_code(tag)
        return {
            "p": self._state._device_id,
            "a": self._state.user.id64,
            "k": code,
            "t": timestamp,
            "m": "android",
            "tag": tag,
        }

    def _assert_valid(self, resp: dict[str, Any]) -> None:
        if not resp.get("success", False):
            self._state._confirmations_to_ignore.append(self.trade_id)
            raise ConfirmationError

    async def _perform_op(self, op: str) -> None:
        params = await self._confirm_params(op)
        params["op"] = op
        params["cid"] = self.data_conf_id
        params["ck"] = self.data_key
        resp = await self._state.http.get(URL.COMMUNITY / "mobileconf/ajaxop", params=params)
        self._assert_valid(resp)

    async def confirm(self) -> None:
        await self._perform_op("allow")

    async def cancel(self) -> None:
        await self._perform_op("cancel")

    async def details(self) -> str:
        params = await self._confirm_params(self.tag)
        resp = await self._state.http.get(URL.COMMUNITY / f"mobileconf/details/{self.id}", params=params)
        self._assert_valid(resp)
        return resp["html"]



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_clientserver_login.proto
# plugin: python-betterproto
# Last updated 19/04/2022

from dataclasses import dataclass
from typing import List

import betterproto


@dataclass(eq=False, repr=False)
class DefinitionBadgeData(betterproto.Message):
    level: int = betterproto.int32_field(1)
    image: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class CommunityItemData(betterproto.Message):
    item_name: str = betterproto.string_field(1)
    item_title: str = betterproto.string_field(2)
    item_description: str = betterproto.string_field(3)
    item_image_small: str = betterproto.string_field(4)
    item_image_large: str = betterproto.string_field(5)
    item_movie_webm: str = betterproto.string_field(6)
    item_movie_mp4: str = betterproto.string_field(7)
    animated: bool = betterproto.bool_field(8)
    badge_data: List[DefinitionBadgeData] = betterproto.message_field(9)
    item_movie_webm_small: str = betterproto.string_field(10)
    item_movie_mp4_small: str = betterproto.string_field(11)
    profile_theme_id: str = betterproto.string_field(12)


@dataclass(eq=False, repr=False)
class Definition(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    defid: int = betterproto.uint32_field(2)
    type: int = betterproto.int32_field(3)
    community_item_class: int = betterproto.int32_field(4)
    community_item_type: int = betterproto.uint32_field(5)
    point_cost: int = betterproto.int64_field(6)
    timestamp_created: int = betterproto.uint32_field(7)
    timestamp_updated: int = betterproto.uint32_field(8)
    timestamp_available: int = betterproto.uint32_field(9)
    quantity: int = betterproto.int64_field(10)
    internal_description: str = betterproto.string_field(11)
    active: bool = betterproto.bool_field(12)
    community_item_data: CommunityItemData = betterproto.message_field(13)
    timestamp_available_end: int = betterproto.uint32_field(14)
    bundle_defids: List[int] = betterproto.uint32_field(15)
    usable_duration: int = betterproto.uint32_field(16)
    bundle_discount: int = betterproto.uint32_field(17)


@dataclass(eq=False, repr=False)
class AddReactionRequest(betterproto.Message):
    target_type: int = betterproto.int32_field(1)
    targetid: int = betterproto.uint64_field(2)
    reactionid: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class AddReactionResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class BatchedQueryRewardItemsRequest(betterproto.Message):
    requests: List["QueryRewardItemsRequest"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class BatchedQueryRewardItemsResponse(betterproto.Message):
    responses: List["BatchedQueryRewardItemsResponseResponse"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class BatchedQueryRewardItemsResponseResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    response: "QueryRewardItemsResponse" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class GetActivePurchaseBonusesRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetActivePurchaseBonusesResponse(betterproto.Message):
    bonuses: List["PurchaseBonus"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetEligibleAppsRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetEligibleAppsResponse(betterproto.Message):
    apps: List["GetEligibleAppsResponseEligibleApp"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetEligibleAppsResponseEligibleApp(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    has_items_anyone_can_purchase: bool = betterproto.bool_field(2)
    event_app: bool = betterproto.bool_field(3)
    hero_carousel_image: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class GetPointsForSpendResponse(betterproto.Message):
    points: int = betterproto.int64_field(1)


@dataclass(eq=False, repr=False)
class GetProfileCustomizationsConfigResponse(betterproto.Message):
    points_cost: int = betterproto.uint32_field(1)
    upgrade_points_cost: int = betterproto.uint32_field(2)
    purchasable_customization_types: List[int] = betterproto.int32_field(3)
    upgradable_customization_types: List[int] = betterproto.int32_field(4)
    max_slots_per_type: int = betterproto.uint32_field(5)
    max_upgradable_level: int = betterproto.uint32_field(6)


@dataclass(eq=False, repr=False)
class GetReactionConfigRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetReactionConfigResponse(betterproto.Message):
    reactions: List["GetReactionConfigResponseReactionConfig"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class GetReactionConfigResponseReactionConfig(betterproto.Message):
    reactionid: int = betterproto.int32_field(1)
    points_cost: int = betterproto.uint32_field(2)
    points_transferred: int = betterproto.uint32_field(3)
    valid_target_types: List[int] = betterproto.int32_field(4)
    valid_ugc_types: List[int] = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class GetReactionsRequest(betterproto.Message):
    target_type: int = betterproto.int32_field(1)
    targetid: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class GetReactionsResponse(betterproto.Message):
    reactionids: List[int] = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class GetReactionsSummaryForUserResponse(betterproto.Message):
    total: List["GetReactionsSummaryForUserResponseBreakdown"] = betterproto.message_field(1)
    user_reviews: List["GetReactionsSummaryForUserResponseBreakdown"] = betterproto.message_field(2)
    ugc: List["GetReactionsSummaryForUserResponseBreakdown"] = betterproto.message_field(3)
    profile: List["GetReactionsSummaryForUserResponseBreakdown"] = betterproto.message_field(4)
    forum_topics: List["GetReactionsSummaryForUserResponseBreakdown"] = betterproto.message_field(5)
    comments: List["GetReactionsSummaryForUserResponseBreakdown"] = betterproto.message_field(6)
    total_points_given: int = betterproto.int64_field(7)
    total_points_received: int = betterproto.int64_field(8)
    total_points_given: int = betterproto.int64_field(9)
    total_points_received: int = betterproto.int64_field(10)


@dataclass(eq=False, repr=False)
class GetReactionsSummaryForUserResponseBreakdown(betterproto.Message):
    reactionid: int = betterproto.int32_field(1)
    given: int = betterproto.uint32_field(2)
    received: int = betterproto.uint32_field(3)
    points_given: int = betterproto.int64_field(4)
    points_received: int = betterproto.int64_field(5)


@dataclass(eq=False, repr=False)
class GetSummaryRequest(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class GetSummaryResponse(betterproto.Message):
    summary: "GetSummaryResponseSummary" = betterproto.message_field(1)
    timestamp_updated: int = betterproto.uint32_field(2)
    auditid_highwater: int = betterproto.uint64_field(3)


@dataclass(eq=False, repr=False)
class GetSummaryResponseSummary(betterproto.Message):
    points: int = betterproto.int64_field(1)
    points_earned: int = betterproto.int64_field(2)
    points_spent: int = betterproto.int64_field(3)


@dataclass(eq=False, repr=False)
class QueryRewardItemsRequest(betterproto.Message):
    appids: List[int] = betterproto.uint32_field(1)
    time_available: int = betterproto.uint32_field(2)
    community_item_classes: List[int] = betterproto.int32_field(3)
    language: str = betterproto.string_field(4)
    count: int = betterproto.int32_field(5)
    cursor: str = betterproto.string_field(6)
    sort: int = betterproto.int32_field(7)
    sort_descending: bool = betterproto.bool_field(8)
    reward_types: List[int] = betterproto.int32_field(9)
    excluded_community_item_classes: List[int] = betterproto.int32_field(10)
    definitionids: List[int] = betterproto.uint32_field(11)
    filters: List[int] = betterproto.int32_field(12)
    filter_match_all_category_tags: List[str] = betterproto.string_field(13)
    filter_match_any_category_tags: List[str] = betterproto.string_field(14)
    contains_definitionids: List[int] = betterproto.uint32_field(15)
    include_direct_purchase_disabled: bool = betterproto.bool_field(16)
    excluded_content_descriptors: List[int] = betterproto.uint32_field(17)
    excluded_appids: List[int] = betterproto.uint32_field(18)


@dataclass(eq=False, repr=False)
class QueryRewardItemsResponse(betterproto.Message):
    definitions: List["Definition"] = betterproto.message_field(1)
    total_count: int = betterproto.int32_field(2)
    count: int = betterproto.int32_field(3)
    next_cursor: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class RedeemPointsRequest(betterproto.Message):
    defid: int = betterproto.uint32_field(1)
    expected_points_cost: int = betterproto.int64_field(2)


@dataclass(eq=False, repr=False)
class RedeemPointsResponse(betterproto.Message):
    communityitemid: int = betterproto.uint64_field(1)
    bundle_community_item_ids: List[int] = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class RedeemPointsForBadgeLevelRequest(betterproto.Message):
    defid: int = betterproto.uint32_field(1)
    num_levels: int = betterproto.int32_field(2)


@dataclass(eq=False, repr=False)
class RedeemPointsForProfileCustomizationRequest(betterproto.Message):
    customization_type: int = betterproto.int32_field(1)


@dataclass(eq=False, repr=False)
class RedeemPointsForProfileCustomizationResponse(betterproto.Message):
    purchaseid: int = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class RedeemPointsForProfileCustomizationUpgradeRequest(betterproto.Message):
    customization_type: int = betterproto.int32_field(1)
    new_level: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class RedeemPointsForProfileCustomizationUpgradeResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class RegisterForSteamDeckRewardsRequest(betterproto.Message):
    serial_number: str = betterproto.string_field(1)
    controller_code: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class RegisterForSteamDeckRewardsResponse(betterproto.Message):
    granted_profile_modifier: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class PurchaseBonus(betterproto.Message):
    bonusid: int = betterproto.uint64_field(1)
    appid: int = betterproto.uint32_field(2)
    active: bool = betterproto.bool_field(3)
    points: int = betterproto.int32_field(4)
    timestamp_start: int = betterproto.uint32_field(5)
    timestamp_end: int = betterproto.uint32_field(6)
    internal_description: str = betterproto.string_field(7)



from typing import NamedTuple

from typing_extensions import Final, Literal

__all__ = (
    "__title__",
    "__author__",
    "__license__",
    "__version__",
    "version_info",
)


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: Literal["alpha", "beta", "candidate", "final"]


__title__: Final = "steam"
__author__: Final = "Gobot1234"
__license__: Final = "MIT"
__version__: Final = "0.9.1"
version_info: Final = VersionInfo(major=0, minor=9, micro=1, releaselevel="final")



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
from collections.abc import Sequence
from typing import TYPE_CHECKING

from .channel import GroupChannel
from .chat import ChatGroup, Member
from .enums import Type

if TYPE_CHECKING:
    from .protobufs import chat
    from .state import ConnectionState
    from .types.id import ChatGroupID
    from .user import User


__all__ = ("Group", "GroupMember")


class GroupMember(Member):
    group: Group
    clan: None

    def __init__(self, state: ConnectionState, group: Group, user: User, proto: chat.Member):
        super().__init__(state, group, user, proto)
        self.group = group


class Group(ChatGroup[GroupMember, GroupChannel]):
    """Represents a Steam group."""

    __slots__ = ()

    def __init__(self, state: ConnectionState, id: ChatGroupID):
        super().__init__(id, type=Type.Chat)
        self._state = state
        self._id = id

    async def chunk(self) -> Sequence[GroupMember]:
        self._members = dict.fromkeys(self._partial_members)  # type: ignore
        for id, member in self._partial_members.items():
            user = self._state.get_user(id)
            while user is None:
                await asyncio.sleep(0)  # same as in clan.py, let receive() populate the cache
                user = self._state.get_user(id)
            member = GroupMember(self._state, self, user, member)
            self._members[member.id] = member
        return await super().chunk()

    # TODO is this possible
    # async def join(self) -> None:
    #     ...



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Generic, TypeVar

from typing_extensions import Never

from .abc import Awardable

if TYPE_CHECKING:
    from .abc import Commentable
    from .message import Authors
    from .reaction import Award, AwardReaction
    from .state import ConnectionState


__all__ = ("Comment",)

OwnerT = TypeVar("OwnerT", bound="Commentable")


@dataclass(repr=False)
class Comment(Awardable, Generic[OwnerT]):
    """Represents a comment on a Steam profile.

    .. container:: operations

        .. describe:: x == y

            Checks if two comments are equal.

        .. describe:: hash(x)

            Returns the comment's hash.

    Attributes
    ----------
    id
        The comment's ID.
    content
        The comment's content.
    author
        The author of the comment.
    created_at
        The time the comment was posted at.
    owner
        The comment sections owner.
    reactions
        The comment's reactions.
    """

    _AWARDABLE_TYPE = 5
    __slots__ = ("_state", "content", "id", "created_at", "reactions", "author", "owner")

    _state: ConnectionState
    id: int
    content: str
    created_at: datetime
    reactions: list[AwardReaction]
    author: Authors
    owner: OwnerT

    def __repr__(self) -> str:
        attrs = ("id", "author")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<Comment {' '.join(resolved)}>"

    def __eq__(self, other: object) -> bool:
        return self.id == other.id if isinstance(other, Comment) else NotImplemented

    def __hash__(self) -> int:
        return hash(self.id)

    async def report(self) -> None:
        """Reports the comment."""
        await self._state.report_comment(self.owner, self.id)

    async def delete(self) -> None:
        """Deletes the comment."""
        await self._state.delete_comment(self.owner, self.id)

    if TYPE_CHECKING:
        # @overload
        # async def award(self: Comment[Topic], award: Award) -> None: ...
        async def award(self, award: Award) -> Never:
            ...



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from collections.abc import Callable, Generator, Mapping
from enum import Enum as _Enum, EnumMeta as _EnumMeta, IntEnum as _IntEnum
from types import MappingProxyType, new_class
from typing import TYPE_CHECKING, Any, Generic, NoReturn, TypeVar, cast

from typing_extensions import Final, Literal, Self

from ._const import DOCS_BUILDING

__all__ = (
    "Enum",
    "IntEnum",
    "Flags",
    "Result",
    "Language",
    "Universe",
    "Type",
    "TypeChar",
    "InstanceFlag",
    "FriendRelationship",
    "PersonaState",
    "PersonaStateFlag",
    "CommunityVisibilityState",
    "TradeOfferState",
    "ChatMemberRank",
    "ChatEntryType",
    "UIMode",
    "UserBadge",
    "ReviewType",
    "GameServerRegion",
    "EventType",
    "ClanEvent",
    "ProfileItemType",
    "ProfileCustomisationStyle",
    "ProfileItemClass",
    "DepotFileFlag",
    "AppFlag",
    "LicenseFlag",
    "LicenseType",
    "BillingType",
    "PaymentMethod",
    "PackageStatus",
    "PublishedFileRevision",
    "PublishedFileType",
    "PublishedFileVisibility",
    "PublishedFileQueryType",
    "PublishedFileQueryFileType",
)

T = TypeVar("T")

T_co = TypeVar("T_co")
TT_co = TypeVar("TT_co", bound="type[Any]")
# if sys.version_info >= (3, 9):
#     classproperty = lambda func: classmethod(property(func))


class classproperty(Generic[TT_co, T_co]):
    def __init__(self, func: Callable[[TT_co], T_co]):
        self.__func__ = func

    def __get__(self, instance: Any, type: TT_co) -> T_co:
        return self.__func__(type)


def _is_descriptor(obj: object) -> bool:
    """Returns True if obj is a descriptor, False otherwise."""
    return hasattr(obj, "__get__") or hasattr(obj, "__set__") or hasattr(obj, "__delete__")


class EnumMeta(_EnumMeta if TYPE_CHECKING else type):
    _value_map_: Mapping[Any, Enum]
    _member_map_: Mapping[str, Enum]

    def __new__(mcs, name: str, bases: tuple[type, ...], attrs: dict[str, Any]) -> Self:
        enum_class: type[Enum] = super().__new__(mcs, name, bases, attrs)  # type: ignore

        value_mapping: dict[Any, Enum] = {}
        member_mapping: dict[str, Enum] = {}

        for key, value in attrs.items():
            if key[0] == "_" or _is_descriptor(value):
                continue

            member = value_mapping.get(value)
            if member is None:
                member = enum_class.__new__(enum_class, name=key, value=value)
                value_mapping[member.value] = member

            member_mapping[key] = member
            type.__setattr__(enum_class, key, member)

        type.__setattr__(enum_class, "_value_map_", value_mapping)
        type.__setattr__(enum_class, "_member_map_", member_mapping)
        return enum_class

    if not TYPE_CHECKING:

        def __call__(cls, value: Any) -> Enum:
            try:
                return cls._value_map_[value]
            except (KeyError, TypeError):
                raise ValueError(f"{value!r} is not a valid {cls.__name__}")

        def __iter__(cls) -> Generator[Enum, None, None]:
            yield from cls._member_map_.values()

        def __reversed__(cls) -> Generator[Enum, None, None]:
            yield from reversed(tuple(cls._member_map_.values()))  # can remove tuple cast after 3.7

        def __getitem__(cls, key: str) -> Enum:
            return cls._member_map_[key]

        @property
        def __members__(cls) -> MappingProxyType[str, Enum]:
            return MappingProxyType(cls._member_map_)

    def __repr__(cls) -> str:
        return f"<enum {cls.__name__!r}>"

    def __len__(cls) -> int:
        return len(cls._member_map_)

    def __setattr__(cls, name: str, value: Any) -> NoReturn:
        raise AttributeError(f"{cls.__name__}: cannot reassign Enum members.")

    def __delattr__(cls, name: str) -> NoReturn:
        raise AttributeError(f"{cls.__name__}: cannot delete Enum members.")

    def __contains__(cls, member: object) -> bool:
        return (
            isinstance(member, cls) and member.name in cls._member_map_ if isinstance(member, Enum) else NotImplemented
        )


# pretending these are enum subclasses makes things much nicer for linters as enums have custom behaviour you can't
# replicate in the current type system
class Enum(_Enum if TYPE_CHECKING else object, metaclass=EnumMeta):
    """A general enumeration, emulates `enum.Enum`."""

    name: str
    value: Any

    def __new__(cls, *, name: str, value: Any) -> Self:
        # N.B. this method is not ever called after enum creation as it is shadowed by EnumMeta.__call__ and is just
        # for creating Enum members
        super_ = super()
        self = (
            super_.__new__(cls, value)
            if any(not issubclass(base, Enum) for base in cls.__mro__[:-1])  # is it is a mixin enum
            else super_.__new__(cls)  # type: ignore
        )
        super_.__setattr__(self, "name", name)
        super_.__setattr__(self, "value", value)
        return self

    def __setattr__(self, key: str, value: Any) -> NoReturn:
        raise AttributeError(f"{self.__class__.__name__} Cannot reassign an Enum members attribute's.")

    def __delattr__(self, item: Any) -> NoReturn:
        raise AttributeError(f"{self.__class__.__name__} Cannot delete an Enum members attribute's.")

    def __bool__(self) -> Literal[True]:
        return True  # an enum member with a zero value would return False otherwise

    def __str__(self) -> str:
        return f"{self.__class__.__name__}.{self.name}"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}.{self.name}: {self.value!r}>"

    @classmethod
    def try_value(cls: type[Self], value: Any) -> Self:
        try:
            return cls._value_map_[value]
        except (KeyError, TypeError):
            return cls.__new__(cls, name=f"{cls.__name__}UnknownValue", value=value)


class IntEnum(Enum, int):
    """An enumeration where all the values are integers, emulates `enum.IntEnum`."""


if cast(Literal[False], DOCS_BUILDING):
    enum_dict = Enum.__dict__.copy()
    for key in ("__new__", "__setattr__"):
        del enum_dict[key]

    Enum = new_class("Enum", (_Enum,), {}, lambda ns: ns.update(enum_dict))  # type: ignore
    IntEnum = new_class("IntEnum", (Enum, _IntEnum))  # type: ignore


class Flags(IntEnum):
    @classmethod
    def try_value(cls, value: int) -> Self:
        if value == 0:
            # causes flags to be iter(cls) does |= on every value (which is not 0) it always returns UnknownValue
            return super().try_value(value)
        flags = (enum for enum in cls if enum.value & value)
        returning_flag = next(flags, None)
        if returning_flag is not None:
            for flag in flags:
                returning_flag |= flag
            if returning_flag == value:
                return returning_flag
        return cls.__new__(cls, name=f"{cls.__name__}UnknownValue", value=value)

    def __or__(self, other: Self | int) -> Self:
        cls = self.__class__
        value = self.value | int(other)
        try:
            return cls._value_map_[value]
        except KeyError:
            return cls.__new__(cls, name=f"{self.name} | {getattr(other, 'name', other)}", value=value)

    def __and__(self, other: Self | int) -> Self:
        cls = self.__class__
        value = self.value & int(other)
        try:
            return cls._value_map_[value]
        except KeyError:
            return cls.__new__(cls, name=f"{self.name} & {getattr(other, 'name', other)}", value=value)

    __ror__ = __or__
    __rand__ = __and__


# fmt: off
class Result(IntEnum):
    # these are a combination of https://partner.steamgames.com/doc/api/steam_api#EResult and https://steamerrors.com
    Invalid                         = 0  #: Invalid Result.
    OK                              = 1  #: Success.
    Fail                            = 2  #: Generic failure.
    NoConnection                    = 3  #: Your Steam client doesn't have a connection to the back-end.
    InvalidPassword                 = 5  #: Password/ticket is invalid.
    LoggedInElsewhere               = 6  #: Same user logged in elsewhere.
    InvalidProtocolVersion          = 7  #: Protocol version is incorrect.
    InvalidParameter                = 8  #: A parameter is incorrect.
    FileNotFound                    = 9  #: File was not found.
    Busy                            = 10  #: Called method busy - action not taken.
    InvalidState                    = 11  #: Called object was in an invalid state.
    InvalidName                     = 12  #: The name was invalid.
    InvalidEmail                    = 13  #: The email was invalid.
    DuplicateName                   = 14  #: The name is not unique.
    AccessDenied                    = 15  #: Access is denied.
    Timeout                         = 16  #: Operation timed out.
    Banned                          = 17  #: VAC2 banned.
    AccountNotFound                 = 18  #: Account not found.
    InvalidSteamID                  = 19  #: The Steam ID was invalid.
    ServiceUnavailable              = 20  #: The requested service is currently unavailable.
    NotLoggedOn                     = 21  #: The user is not logged on.
    Pending                         = 22  #: Request is pending (may be in process, or waiting on third party).
    EncryptionFailure               = 23  #: Encryption or decryption failed.
    InsufficientPrivilege           = 24  #: Insufficient privilege.
    LimitExceeded                   = 25  #: Too much of a good thing.
    Revoked                         = 26  #: Access has been revoked (used for revoked guest passes).
    Expired                         = 27  #: License/Guest pass the user is trying to access is expired.
    AlreadyRedeemed                 = 28  #: Guest pass has already been redeemed by account, cannot be acknowledged again.
    DuplicateRequest                = 29  #: The request is a duplicate, ignored this time.
    AlreadyOwned                    = 30  #: All the games in guest pass redemption request are already owned by the user.
    IPNotFound                      = 31  #: IP address not found.
    PersistFailed                   = 32  #: Failed to write change to the data store.
    LockingFailed                   = 33  #: Failed to acquire access lock for this operation.
    LogonSessionReplaced            = 34  #: The logon session has been replaced.
    ConnectFailed                   = 35  #: Failed to connect.
    HandshakeFailed                 = 36  #: The authentication handshake has failed.
    IOFailure                       = 37  #: Generic IO failure.
    RemoteDisconnect                = 38  #: The remote server has disconnected.
    ShoppingCartNotFound            = 39  #: Failed to find the shopping cart requested.
    Blocked                         = 40  #: A user blocked the action.
    Ignored                         = 41  #: The target is ignoring sender.
    NoMatch                         = 42  #: Nothing matching the request found.
    AccountDisabled                 = 43  #: The account is disabled.
    ServiceReadOnly                 = 44  #: This service is not accepting content changes right now.
    AccountNotFeatured              = 45  #: Account doesn't have value, so this feature isn't available.
    AdministratorOK                 = 46  #: Allowed to take this action, but only because requester is admin.
    ContentVersion                  = 47  #: A Version mismatch in content transmitted within the Steam protocol.
    TryAnotherCM                    = 48  #: The current CM can't service the user making a request, should try another.
    PasswordRequiredToKickSession   = 49  #: You are already logged in elsewhere, this cached credential login has failed.
    AlreadyLoggedInElsewhere        = 50  #: You are already logged in elsewhere, you must wait.
    Suspended                       = 51  #: Long running operation (content download) suspended/paused.
    Cancelled                       = 52  #: Operation canceled (typically by user content download).
    DataCorruption                  = 53  #: Operation canceled because data is malformed or unrecoverable.
    DiskFull                        = 54  #: Operation canceled - not enough disk space.
    RemoteCallFailed                = 55  #: An remote call or IPC call failed.
    ExternalAccountUnlinked         = 57  #: External account is not linked to a Steam account.
    PSNTicketInvalid                = 58  #: PSN ticket was invalid.
    ExternalAccountAlreadyLinked    = 59  #: External account is already linked to some other account.
    RemoteFileConflict              = 60  #: The sync cannot resume due to a conflict between the local and remote files.
    IllegalPassword                 = 61  #: The requested new password is not legal.
    SameAsPreviousValue             = 62  #: New value is the same as the old one (secret question and answer).
    AccountLogonDenied              = 63  #: Account login denied due to 2nd factor authentication failure.
    CannotUseOldPassword            = 64  #: The requested new password is not legal.
    InvalidLoginAuthCode            = 65  #: Account login denied due to auth code invalid.
    AccountLogonDeniedNoMail        = 66  #: Account login denied due to 2nd factor authentication failure.
    HardwareNotCapableOfIPT         = 67  #: The user's hardware does not support Intel's identity protection technology.
    IPTInitError                    = 68  #: Intel's Identity Protection Technology has failed to initialize.
    ParentalControlRestricted       = 69  #: Operation failed due to parental control restrictions for current user.
    FacebookQueryError              = 70  #: Facebook query returned an error.
    ExpiredLoginAuthCode            = 71  #: Account login denied due to auth code expired.
    IPLoginRestrictionFailed        = 72  #: The login failed due to an IP restriction.
    AccountLockedDown               = 73  #: The current users account is currently locked for use.
    VerifiedEmailRequired           = 74  #: The logon failed because the accounts email is not verified.
    NoMatchingURL                   = 75  #: There is no url matching the provided values.
    BadResponse                     = 76  #: Parse failure, missing field, etc.
    RequirePasswordReEntry          = 77  #: The user cannot complete the action until they re-enter their password.
    ValueOutOfRange                 = 78  #: The value entered is outside the acceptable range.
    UnexpectedError                 = 79  #: Something happened that we didn't expect to ever happen.
    Disabled                        = 80  #: The requested service has been configured to be unavailable.
    InvalidCEGSubmission            = 81  #: The set of files submitted to the CEG server are not valid.
    RestrictedDevice                = 82  #: The device being used is not allowed to perform this action.
    RegionLocked                    = 83  #: The action could not be complete because it is region restricted.
    RateLimitExceeded               = 84  #: Temporary rate limit exceeded. Different from :attr:`LimitExceeded`.
    LoginDeniedNeedTwoFactor        = 85  #: Need two-factor code to log in.
    ItemDeleted                     = 86  #: The thing we're trying to access has been deleted.
    AccountLoginDeniedThrottle      = 87  #: Login attempt failed, try to throttle response to possible attacker.
    TwoFactorCodeMismatch           = 88  #: Two-factor code mismatch.
    TwoFactorActivationCodeMismatch = 89  #: Activation code for two-factor didn't match.
    NotModified                     = 91  #: Data not modified.
    TimeNotSynced                   = 93  #: The time presented is out of range or tolerance.
    SMSCodeFailed                   = 94  #: SMS code failure (no match, none pending, etc.).
    AccountActivityLimitExceeded    = 96  #: Too many changes to this account.
    PhoneActivityLimitExceeded      = 97  #: Too many changes to this phone.
    RefundToWallet                  = 98  #: Cannot refund to payment method, must use wallet.
    EmailSendFailure                = 99  #: Cannot send an email.
    NotSettled                      = 100  #: Can't perform operation till payment has settled.
    NeedCaptcha                     = 101  #: Needs to provide a valid captcha.
    GSLTDenied                      = 102  #: A game server login token owned by this token's owner has been banned.
    GSOwnerDenied                   = 103  #: Game server owner is denied for other reason.
    InvalidItemType                 = 104  #: The type of thing we were requested to act on is invalid.
    IPBanned                        = 105  #: The IP address has been banned from taking this action.
    GSLTExpired                     = 106  #: This Game Server Login Token has expired from disuse; can be reset for use.
    InsufficientFunds               = 107  #: User doesn't have enough wallet funds to complete the action.
    TooManyPending                  = 108  #: There are too many of this thing pending already.
    NoSiteLicensesFound             = 109  #: No site licenses found.
    WGNetworkSendExceeded           = 110  #: The WG couldn't send a response because we exceeded max network send size.
    AccountNotFriends               = 111  #: Not friends with the relevant account.
    LimitedUserAccount              = 112  #: The account is limited and cannot perform this action.
    CantRemoveItem                  = 113  #: Cannot remove the item.
    AccountHasBeenDeleted           = 114  #: The relevant account has been deleted.
    AccountHasCancelledLicense      = 115  #: The user has a user cancelled license.
    DeniedDueToCommunityCooldown    = 116  #: The request was denied due to community cooldown.
    NoLauncherSpecified             = 117  #: No launcher was specified.
    MustAgreeToSSA                  = 118  #: User must agree to China SSA or global SSA before login.
    LauncherMigrated                = 119  #: The specified launcher type is no longer supported.
    SteamRealmMismatch              = 120  #: The user's realm does not match the realm of the requested resource.
    InvalidSignature                = 121  #: Signature check did not match.
    ParseFailure                    = 122  #: Failed to parse input.
    NoVerifiedPhone                 = 123  #: Account does not have a verified phone number.
    InsufficientBatteryCharge       = 124  #: The device battery is too low to complete the action.


class Language(IntEnum):
    NONE                        = -1
    English                     = 0
    German                      = 1
    French                      = 2
    Italian                     = 3
    Korean                      = 4
    Spanish                     = 5
    SimplifiedChinese           = 6
    TraditionalChinese          = 7
    Russian                     = 8
    Thai                        = 9
    Japanese                    = 10
    Portuguese                  = 11
    Polish                      = 12
    Danish                      = 13
    Dutch                       = 14
    Finnish                     = 15
    Norwegian                   = 16
    Swedish                     = 17
    Romanian                    = 18
    Turkish                     = 19
    Hungarian                   = 20
    Czech                       = 21
    PortugueseBrazil            = 22
    Bulgarian                   = 23
    Greek                       = 24
    Arabic                      = 25
    Ukrainian                   = 26
    SpanishLatinAmerican        = 27
    Vietnamese                  = 28
    # SteamChinaSimplifiedChinese = 29  # not including until this appears on Steamworks

    @classproperty
    def NATIVE_NAME_MAP(cls: type[Language]) -> Mapping[Language, str]:  # type: ignore
        """A mapping of every language's native name."""
        return {
            cls.Arabic:               "العربية",
            cls.Bulgarian:            "езбългарски езикик",
            cls.SimplifiedChinese:    "简体中文",
            cls.TraditionalChinese:   "繁體中文",
            cls.Czech:                "čeština",
            cls.Danish:               "Dansk",
            cls.Dutch:                "Nederlands",
            cls.English:              "English",
            cls.Finnish:              "Suomi",
            cls.French:               "Français",
            cls.German:               "Deutsch",
            cls.Greek:                "Ελληνικά",
            cls.Hungarian:            "Magyar",
            cls.Italian:              "Italiano",
            cls.Japanese:             "日本語",
            cls.Korean:               "한국어",
            cls.Norwegian:            "Norsk",
            cls.Polish:               "Polski",
            cls.Portuguese:           "Português",
            cls.PortugueseBrazil:     "Português-Brasil",
            cls.Romanian:             "Română",
            cls.Russian:              "Русский",
            cls.Spanish:              "Español-España",
            cls.SpanishLatinAmerican: "Español-Latinoamérica",
            cls.Swedish:              "Svenska",
            cls.Thai:                 "ไทย",
            cls.Turkish:              "Türkçe",
            cls.Ukrainian:            "Українська",
            cls.Vietnamese:           "Tiếng Việt",
        }

    @classproperty
    def API_LANGUAGE_MAP(cls: type[Language]) -> Mapping[Language, str]:  # type: ignore
        """A mapping of every language's name from the Steamworks API."""
        return {
            cls.Arabic:               "arabic",
            cls.Bulgarian:            "bulgarian",
            cls.SimplifiedChinese:    "schinese",
            cls.TraditionalChinese:   "tchinese",
            cls.Czech:                "czech",
            cls.Danish:               "danish",
            cls.Dutch:                "dutch",
            cls.English:              "english",
            cls.Finnish:              "finnish",
            cls.French:               "french",
            cls.German:               "german",
            cls.Greek:                "greek",
            cls.Hungarian:            "hungarian",
            cls.Italian:              "italian",
            cls.Japanese:             "japanese",
            cls.Korean:               "koreana",
            cls.Norwegian:            "norwegian",
            cls.Polish:               "polish",
            cls.Portuguese:           "portuguese",
            cls.PortugueseBrazil:     "brazilian",
            cls.Romanian:             "romanian",
            cls.Russian:              "russian",
            cls.Spanish:              "spanish",
            cls.SpanishLatinAmerican: "latam",
            cls.Swedish:              "swedish",
            cls.Thai:                 "thai",
            cls.Turkish:              "turkish",
            cls.Ukrainian:            "ukrainian",
            cls.Vietnamese:           "vietnamese",
        }

    @classproperty
    def WEB_API_MAP(cls: type[Language]) -> Mapping[Language, str]:  # type: ignore
        """A mapping of every language's name from the Web API."""
        return {
            cls.Arabic:               "ar",
            cls.Bulgarian:            "bg",
            cls.SimplifiedChinese:    "zh-CN",
            cls.TraditionalChinese:   "zh-TW",
            cls.Czech:                "cs",
            cls.Danish:               "da",
            cls.Dutch:                "nl",
            cls.English:              "en",
            cls.Finnish:              "fi",
            cls.French:               "fr",
            cls.German:               "de",
            cls.Greek:                "el",
            cls.Hungarian:            "hu",
            cls.Italian:              "it",
            cls.Japanese:             "ja",
            cls.Korean:               "ko",
            cls.Norwegian:            "no",
            cls.Polish:               "pl",
            cls.Portuguese:           "pt",
            cls.PortugueseBrazil:     "pt-BR",
            cls.Romanian:             "ro",
            cls.Russian:              "ru",
            cls.Spanish:              "es",
            cls.SpanishLatinAmerican: "es-419",
            cls.Swedish:              "sv",
            cls.Thai:                 "th",
            cls.Turkish:              "tr",
            cls.Ukrainian:            "uk",
            cls.Vietnamese:           "vn",
        }

    @property
    def native_name(self) -> str:
        """This language's native name."""
        return self.NATIVE_NAME_MAP[self]

    @property
    def api_name(self) -> str:
        """This language's Steamworks name."""
        return self.API_LANGUAGE_MAP[self]

    @property
    def web_api_name(self) -> str:
        """This language's Web API name."""
        return self.WEB_API_MAP[self]

    @classmethod
    def from_str(cls, string: str) -> Self:
        try:
            return _REVERSE_API_LANGUAGE_MAP[string.lower()]
        except KeyError:
            return cls.__new__(cls, name=string.title(), value=-1)


_REVERSE_API_LANGUAGE_MAP: Final = cast(
    "Mapping[str, Language]", {value: key for key, value in Language.API_LANGUAGE_MAP.items()}
)


class Universe(IntEnum):
    Invalid  = 0  #: Invalid.
    Public   = 1  #: The standard public universe.
    Beta     = 2  #: Beta universe used inside Valve.
    Internal = 3  #: Internal universe used inside Valve.
    Dev      = 4  #: Dev universe used inside Valve.
    Max      = 6  #: Total number of universes, used for sanity checks.


class Type(IntEnum):
    Invalid        = 0   #: Used for invalid Steam IDs.
    Individual     = 1   #: Single user account.
    Multiseat      = 2   #: Multiseat (e.g. cybercafe) account.
    GameServer     = 3   #: Game server account.
    AnonGameServer = 4   #: Anonymous game server account.
    Pending        = 5   #: Pending.
    ContentServer  = 6   #: Valve internal content server account.
    Clan           = 7   #: Steam clan.
    Chat           = 8   #: Steam group chat or lobby.
    ConsoleUser    = 9   #: Fake SteamID for local PSN account on PS3 or Live account on 360, etc.
    AnonUser       = 10  #: Anonymous user account. (Used to create an account or reset a password)
    Max            = 11  #: Max of 16 items in this field


class TypeChar(IntEnum):
    I = Type.Invalid         #: The character used for :class:`~steam.Type.Invalid`.
    U = Type.Individual      #: The character used for :class:`~steam.Type.Individual`.
    M = Type.Multiseat       #: The character used for :class:`~steam.Type.Multiseat`.
    G = Type.GameServer      #: The character used for :class:`~steam.Type.GameServer`.
    A = Type.AnonGameServer  #: The character used for :class:`~steam.Type.AnonGameServer`.
    P = Type.Pending         #: The character used for :class:`~steam.Type.Pending`.
    C = Type.ContentServer   #: The character used for :class:`~steam.Type.ContentServer`.
    g = Type.Clan            #: The character used for :class:`~steam.Type.Clan`.
    T = Type.Chat            #: The character used for :class:`~steam.Type.Chat` (Lobby/group chat).
    L = Type.Chat            #: The character used for :class:`~steam.Type.Chat` (Lobby/group chat).
    c = Type.Clan            #: The character used for :class:`~steam.Type.Clan`.
    a = Type.AnonUser        #: The character used for :class:`~steam.Type.AnonUser`.


class InstanceFlag(Flags):
    All          = 0         #: The Instance for all Steam IDs
    Desktop      = 1 << 0    #: The Instance for desktop Steam IDs
    Console      = 1 << 1    #: The Instance for console  Steam IDs
    Web          = 1 << 2    #: The Instance for web Steam IDs
    # I have observed these flags for game servers, but I don't know what they are
    Unknown1     = 1 << 3
    Unknown2     = 1 << 4
    Unknown3     = 1 << 5
    Unknown4     = 1 << 6
    Unknown5     = 1 << 7
    Unknown6     = 1 << 8
    Unknown7     = 1 << 9
    Unknown8     = 1 << 10
    Unknown9     = 1 << 11
    Unknown10    = 1 << 12
    Unknown11    = 1 << 13
    Unknown12    = 1 << 14
    # Unknown13    = 1 << 15
    # Unknown14    = 1 << 16
    # 20474 ->   | 2 |   | 8 | 16 | 32 | 64 | 128 | 256 | 512 | 1024 | 2048 | 16384
    # 20475 -> 1 | 2 |   | 8 | 16 | 32 | 64 | 128 | 256 | 512 | 1024 | 2048 | 16384
    # 20476 ->   |   | 4 | 8 | 16 | 32 | 64 | 128 | 256 | 512 | 1024 | 2048 | 16384
    # Type.Chat exclusive flags
    ChatMMSLobby = 1 << 17  #: The Steam ID is for an MMS Lobby.
    ChatLobby    = 1 << 18  #: The Steam ID is for a Lobby.
    ChatClan     = 1 << 19  #: The Steam ID is for a Clan.


class FriendRelationship(IntEnum):
    NONE             = 0  #: The user has no relationship to you.
    Blocked          = 1  #: The user has been blocked.
    RequestRecipient = 2  #: The user has requested to be friends with you.
    Friend           = 3  #: The user is friends with you.
    RequestInitiator = 4  #: You have requested to be friends with the user.
    Ignored          = 5  #: You have explicitly blocked this other user from comments/chat/etc.
    IgnoredFriend    = 6  #: The user has ignored the current user.
    Max              = 8  #: The total number of friend relationships used for looping and verification.


class PersonaState(IntEnum):
    Offline        = 0  #: The user is not currently logged on.
    Online         = 1  #: The user is logged on.
    Busy           = 2  #: The user is on, but busy.
    Away           = 3  #: The user has been marked as AFK for a short period of time.
    Snooze         = 4  #: The user has been marked as AFK for a long period of time.
    LookingToTrade = 5  #: The user is online and wanting to trade.
    LookingToPlay  = 6  #: The user is online and wanting to play.
    Invisible      = 7  #: The user is invisible.
    Max            = 8  #: The total number of states. Only used for looping and validation.


class PersonaStateFlag(Flags):
    NONE                 = 0
    HasRichPresence      = 1 << 0
    InJoinableGame       = 1 << 1
    Golden               = 1 << 2
    RemotePlayTogether   = 1 << 3
    ClientTypeWeb        = 1 << 4
    ClientTypeMobile     = 1 << 8
    ClientTypeTenfoot    = 1 << 10
    ClientTypeVR         = 1 << 11
    LaunchTypeGamepad    = 1 << 12
    LaunchTypeCompatTool = 1 << 13


class CommunityVisibilityState(IntEnum):
    NONE        = 0  #: The user has no community state.
    Private     = 1  #: The user has a private profile.
    FriendsOnly = 2  #: The user has a friends only profile.
    Public      = 3  #: The user has a public profile.


class TradeOfferState(IntEnum):
    Invalid                   = 1   #: The trade offer's state is invalid.
    Active                    = 2   #: The trade offer is active.
    Accepted                  = 3   #: The trade offer has been accepted.
    Countered                 = 4   #: The trade offer has been countered.
    Expired                   = 5   #: The trade offer has expired.
    Canceled                  = 6   #: The trade offer has been cancelled.
    Declined                  = 7   #: The trade offer has be declined by the partner.
    InvalidItems              = 8   #: The trade offer has invalid items and has been cancelled.
    ConfirmationNeed          = 9   #: The trade offer needs confirmation.
    CanceledBySecondaryFactor = 10  #: The trade offer was cancelled by second factor.
    StateInEscrow             = 11  #: The trade offer is in escrow.

    @property
    def event_name(self) -> str | None:
        try:
            return {
                TradeOfferState.Accepted: "accept",
                TradeOfferState.Countered: "counter",
                TradeOfferState.Expired: "expire",
                TradeOfferState.Canceled: "cancel",
                TradeOfferState.Declined: "decline",
                TradeOfferState.CanceledBySecondaryFactor: "cancel",
            }[self]
        except KeyError:
            return None


class ChatMemberRank(IntEnum):
    Default = 0
    Viewer = 10
    Guest = 15
    Member = 20
    Moderator = 30
    Officer = 40
    Owner = 50
    TestInvalid = 99


class ChatEntryType(IntEnum):
    Invalid          = 0   #: An Invalid Chat entry.
    Text             = 1   #: A Normal text message from another user.
    Typing           = 2   #: Another user is typing (not used in multi-user chat).
    InviteGame       = 3   #: An Invite from other user into that users current game.
    LeftConversation = 6   #: A user has left the conversation.
    Entered          = 7   #: A user has entered the conversation (used in multi-user chat and group chat).
    WasKicked        = 8   #: A user was kicked.
    WasBanned        = 9   #: A user was banned.
    Disconnected     = 10  #: A user disconnected.
    HistoricalChat   = 11  #: A chat message from user's chat history or offline message.
    LinkBlocked      = 14  #: A link was removed by the chat filter.


class UIMode(IntEnum):
    Desktop    = 0  #: The UI mode for the desktop client.
    BigPicture = 1  #: The UI mode for big picture mode.
    Mobile     = 2  #: The UI mode for mobile.
    Web        = 3  #: The UI mode for the web client.


class UserBadge(IntEnum):
    Invalid                           = 0  #: Invalid Badge.
    YearsOfService                    = 1  #: The years of service badge.
    Community                         = 2  #: The pillar of the community badge.
    Portal2PotatoARG                  = 3  #: The portal to potato badge.
    TreasureHunt                      = 4  #: The treasure hunter badge.
    SummerSale2011                    = 5  #: The Summer sale badge for 2011.
    WinterSale2011                    = 6  #: The Winter sale badge for 2011.
    SummerSale2012                    = 7  #: The Summer sale badge for 2012.
    WinterSale2012                    = 8  #: The Winter sale badge for 2012.
    CommunityTranslator               = 9  #: The community translator badge.
    CommunityModerator                = 10  #: The community moderator badge.
    ValveEmployee                     = 11  #: The Valve employee badge.
    GameDeveloper                     = 12  #: The game developer badge.
    GameCollector                     = 13  #: The game collector badge.
    TradingCardBetaParticipant        = 14  #: The trading card beta participant badge.
    SteamBoxBeta                      = 15  #: The Steam box beta badge.
    Summer2014RedTeam                 = 16  #: The Summer sale badge for 2014 for the red team.
    Summer2014BlueTeam                = 17  #: The Summer sale badge for 2014 for the blue team.
    Summer2014PinkTeam                = 18  #: The Summer sale badge for 2014 for the pink team.
    Summer2014GreenTeam               = 19  #: The Summer sale badge for 2014 for the green team.
    Summer2014PurpleTeam              = 20  #: The Summer sale badge for 2014 for the purple team.
    Auction2014                       = 21  #: The auction badge for 2014.
    GoldenProfile2014                 = 22  #: The golden profile for 2014.
    TowerAttackMiniGame               = 23  #: The tower attack mini-game badge.
    Winter2015ARGRedHerring           = 24  #: The Winter ARG red herring badge for 2015.
    SteamAwards2016Nominations        = 25  #: The Steam Awards Nominations badge for 2016.
    StickerCompletionist2017          = 26  #: The sticker completionist badge for 2017.
    SteamAwards2017Nominations        = 27  #: The Steam Awards Nominations badge for 2017.
    SpringCleaning2018                = 28  #: The Spring cleaning badge for 2018.
    Salien                            = 29  #: The Salien badge.
    RetiredModerator                  = 30  #: The retired moderator badge.
    SteamAwards2018Nominations        = 31  #: The Steam Awards Nominations badge for 2018.
    ValveModerator                    = 32  #: The Valve moderator badge.
    WinterSale2018                    = 33  #: The Winter sale badge for 2018.
    LunarNewYearSale2019              = 34  #: The lunar new years sale badge for 2019.
    LunarNewYearSale2019GoldenProfile = 35  #: The lunar new year golden profile sale badge for 2019.
    SpringCleaning2019                = 36  #: The Spring cleaning badge for 2019.
    Summer2019                        = 37  #: The Summer sale badge for 2019.
    Summer2019TeamHare                = 38  #: The Summer sale badge for 2014 for team hare.
    Summer2019TeamTortoise            = 39  #: The Summer sale badge for 2014 for team tortoise.
    Summer2019TeamCorgi               = 40  #: The Summer sale badge for 2014 for team corgi.
    Summer2019TeamCockatiel           = 41  #: The Summer sale badge for 2014 for team cockatiel.
    Summer2019TeamPig                 = 42  #: The Summer sale badge for 2014 for team pig.
    SteamAwards2019Nominations        = 43  #: The Steam Awards Nominations badge for 2019.
    WinterSaleEvent2019               = 44  #: The Winter sale badge for 2019.
    WinterSale2019Steamville          = 45  #: The Winter sale Steamville badge for 2019.
    LunarNewYearSale2020              = 46  #: The lunar new years sale badge for 2020.
    SpringCleaning2020                = 47  #: The Spring cleaning badge for 2020.
    AwardsCommunityContributor        = 48  #: The Steam Awards Community Contributor badge.
    AwardsCommunityPatron             = 49  #: The Steam Awards Community Patron badge.
    SteamAwards2020Nominations        = 50  #: The Steam Awards Nominations badge for 2020.


class ReviewType(IntEnum):
    NONE                   = 0  #: No reviews.
    OverwhelminglyNegative = 1  #: 0 - 19% positive reviews and few of them.
    VeryNegative           = 2  #: 0 - 19% positive reviews.
    Negative               = 3  #: 0 - 39% positive reviews.
    MostlyNegative         = 4  #: 20 - 39% positive reviews but few of them.
    Mixed                  = 5  #: 40 - 69% positive reviews.
    MostlyPositive         = 6  #: 70 - 79% positive reviews.
    Positive               = 7  #: 80 - 100% positive reviews but few of them.
    VeryPositive           = 8  #: 94 - 80% positive reviews.
    OverwhelminglyPositive = 9  #: 95 - 100% positive reviews.


class GameServerRegion(IntEnum):
    NONE         = -1   #: No set game region.
    USEastCoast  = 0    #: A server on the USA's East Coast.
    USWestCoast  = 1    #: A server on the USA's West Coast.
    SouthAmerica = 2    #: A server in South America.
    Europe       = 3    #: A server in Europe.
    Asia         = 4    #: A server in Asia.
    Australia    = 5    #: A server in Australia.
    MiddleEast   = 6    #: A server in the Middle East.
    Africa       = 7    #: A server in Africa.
    World        = 255  #: A server somewhere in the world.


class EventType(IntEnum):
    Other                  = 1  #: An unspecified event.
    Game                   = 2  #: A game event.
    Party                  = 3  #: A party event.
    Meeting                = 4  #: An important meeting.
    SpecialCause           = 5  #: An event for a special cause.
    MusicAndArts           = 6  #: A music or art event.
    Sports                 = 7  #: A sporting event.
    Trip                   = 8  #: A clan trip.
    Chat                   = 9  #: A chat event.
    GameRelease            = 10  #: A game release event.
    Broadcast              = 11  #: A broadcast event.
    SmallUpdate            = 12  #: A small update event.
    PreAnnounceMajorUpdate = 13  #: A pre-announcement for a major update event.
    MajorUpdate            = 14  #: A major update event.
    DLCRelease             = 15  #: A dlc release event.
    FutureRelease          = 16  #: A future release event.
    ESportTournamentStream = 17  #: An e-sport tournament stream event.
    DevStream              = 18  #: A developer stream event.
    FamousStream           = 19  #: A famous stream event.
    GameSales              = 20  #: A game sales event.
    GameItemSales          = 21  #: A game item sales event.
    InGameBonusXP          = 22  #: An in game bonus xp event.
    InGameLoot             = 23  #: An in game loot event.
    InGamePerks            = 24  #: An in game perks event.
    InGameChallenge        = 25  #: An in game challenge event.
    InGameContest          = 26  #: An in game contest event.
    IRL                    = 27  #: An in real life event.
    News                   = 28  #: A news event.
    BetaRelease            = 29  #: A beta release event.
    InGameContentRelease   = 30  #: An in game content release event.
    FreeTrial              = 31  #: A free trial event.
    SeasonRelease          = 32  #: A season release event.
    SeasonUpdate           = 33  #: A season update event.
    Crosspost              = 34  #: A cross post event.
    InGameGeneral          = 35  #: An in game general event.


if not DOCS_BUILDING:
    class ClanEvent(IntEnum):
        for _member in EventType:
            locals()[_member.name] = _member.value
        del _member


class ProfileItemType(IntEnum):
    Invalid                   = 0   #: An invalid item type.
    RareAchievementShowcase   = 1   #: A rare achievements showcase.
    GameCollector             = 2   #: A game collector section.
    ItemShowcase              = 3   #: An item showcase.
    TradeShowcase             = 4   #: A trade info showcase.
    Badges                    = 5   #: A badges showcase.
    FavouriteGame             = 6   #: A favourite game section.
    ScreenshotShowcase        = 7   #: A screenshot showcase.
    CustomText                = 8   #: A custom text section.
    FavouriteGroup            = 9   #: A favourite game showcase.
    Recommendation            = 10  #: A review showcase.
    WorkshopItem              = 11  #: A workshop item showcase.
    MyWorkshop                = 12  #: A showcase of a workshop item made by profile's owner.
    ArtworkShowcase           = 13  #: An artwork showcase.
    VideoShowcase             = 14  #: A video showcase.
    Guides                    = 15  #: A guide showcase.
    MyGuides                  = 16  #: A showcase of the profile's owner's guides.
    Achievements              = 17  #: The owner's profile's achievements.
    Greenlight                = 18  #: A greenlight showcase.
    MyGreenlight              = 19  #: A showcase of a greenlight game the profiles' owner has made.
    Salien                    = 20  #: A salien showcase.
    LoyaltyRewardReactions    = 21  #: A loyalty reward showcase.
    SingleArtworkShowcase     = 22  #: A single artwork showcase.
    AchievementsCompletionist = 23  #: An achievements completeionist showcase.


class ProfileCustomisationStyle(IntEnum):
    Default      = 0
    Selected     = 1
    Rarest       = 2
    MostRecent   = 3
    Random       = 4
    HighestRated = 5


class ProfileItemClass(IntEnum):
    Invalid               = 0
    Badge                 = 1
    GameCard              = 2
    ProfileBackground     = 3
    Emoticon              = 4
    BoosterPack           = 5
    Consumable            = 6
    GameGoo               = 7
    ProfileModifier       = 8
    Scene                 = 9
    SalienItem            = 10
    Sticker               = 11
    ChatEffect            = 12
    MiniProfileBackground = 13
    AvatarFrame           = 14
    AnimatedAvatar        = 15
    SteamDeckKeyboardSkin = 16


class DepotFileFlag(Flags):
    File                = 0
    UserConfig          = 1 << 0
    VersionedUserConfig = 1 << 1
    Encrypted           = 1 << 2
    ReadOnly            = 1 << 3
    Hidden              = 1 << 4
    Executable          = 1 << 5
    Directory           = 1 << 6
    CustomExecutable    = 1 << 7
    InstallScript       = 1 << 8
    Symlink             = 1 << 9


TYPE_TRANSFORM_MAP: Final = cast("Mapping[str, str]", {
    "Dlc": "DLC",
})


class AppFlag(Flags):
    Game        = 1 << 0
    Application = 1 << 1
    Tool        = 1 << 2
    Demo        = 1 << 3
    Deprecated  = 1 << 4
    DLC         = 1 << 5
    Guide       = 1 << 6
    Driver      = 1 << 7
    Config      = 1 << 8
    Hardware    = 1 << 9
    Franchise   = 1 << 10
    Video       = 1 << 11
    Plugin      = 1 << 12
    Music       = 1 << 13
    Series      = 1 << 14
    Comic       = 1 << 15
    Beta        = 1 << 16

    Shortcut    = 1 << 30
    DepotOnly   = -1 << 31

    @classmethod
    def from_str(cls, name: str) -> Self:
        types = iter(name.split(","))
        type = next(types).strip().title()
        self = cls[TYPE_TRANSFORM_MAP.get(type, type)]
        for type in types:
            type = type.strip().title()
            self |= cls[TYPE_TRANSFORM_MAP.get(type, type)]
        return self


class LicenseFlag(Flags):
    NONE                         = 0
    Renew                        = 1 << 0
    RenewalFailed                = 1 << 1
    Pending                      = 1 << 2
    Expired                      = 1 << 3
    CancelledByUser              = 1 << 4
    CancelledByAdmin             = 1 << 5
    LowViolenceContent           = 1 << 6
    ImportedFromSteam2           = 1 << 7
    ForceRunRestriction          = 1 << 8
    RegionRestrictionExpired     = 1 << 9
    CancelledByFriendlyFraudLock = 1 << 10
    NotActivated                 = 1 << 11


class LicenseType(IntEnum):
    NoLicense                             = 0
    SinglePurchase                        = 1
    SinglePurchaseLimitedUse              = 2
    RecurringCharge                       = 3
    RecurringChargeLimitedUse             = 4
    RecurringChargeLimitedUseWithOverages = 5
    RecurringOption                       = 6
    LimitedUseDelayedActivation           = 7


class BillingType(IntEnum):
    NoCost                 = 0
    BillOnceOnly           = 1
    BillMonthly            = 2
    ProofOfPrepurchaseOnly = 3
    GuestPass              = 4
    HardwarePromo          = 5
    Gift                   = 6
    AutoGrant              = 7
    OEMTicket              = 8
    RecurringOption        = 9
    BillOnceOrCDKey        = 10
    Repurchaseable         = 11
    FreeOnDemand           = 12
    Rental                 = 13
    CommercialLicense      = 14
    FreeCommercialLicense  = 15
    NumBillingTypes        = 16


class PaymentMethod(IntEnum):
    NONE                   = 0
    ActivationCode         = 1
    CreditCard             = 2
    Giropay                = 3
    PayPal                 = 4
    Ideal                  = 5
    PaySafeCard            = 6
    Sofort                 = 7
    GuestPass              = 8
    WebMoney               = 9
    MoneyBookers           = 10
    AliPay                 = 11
    Yandex                 = 12
    Kiosk                  = 13
    Qiwi                   = 14
    GameStop               = 15
    HardwarePromo          = 16
    MoPay                  = 17
    BoletoBancario         = 18
    BoaCompraGold          = 19
    BancoDoBrasilOnline    = 20
    ItauOnline             = 21
    BradescoOnline         = 22
    Pagseguro              = 23
    VisaBrazil             = 24
    AmexBrazil             = 25
    Aura                   = 26
    Hipercard              = 27
    MastercardBrazil       = 28
    DinersCardBrazil       = 29
    AuthorizedDevice       = 30
    MOLPoints              = 31
    ClickAndBuy            = 32
    Beeline                = 33
    Konbini                = 34
    EClubPoints            = 35
    CreditCardJapan        = 36
    BankTransferJapan      = 37
    PayEasy                = 38
    Zong                   = 39
    CultureVoucher         = 40
    BookVoucher            = 41
    HappymoneyVoucher      = 42
    ConvenientStoreVoucher = 43
    GameVoucher            = 44
    Multibanco             = 45
    Payshop                = 46
    MaestroBoaCompra       = 47
    OXXO                   = 48
    ToditoCash             = 49
    Carnet                 = 50
    SPEI                   = 51
    ThreePay               = 52
    IsBank                 = 53
    Garanti                = 54
    Akbank                 = 55
    YapiKredi              = 56
    Halkbank               = 57
    BankAsya               = 58
    Finansbank             = 59
    DenizBank              = 60
    PTT                    = 61
    CashU                  = 62
    AutoGrant              = 64
    WebMoneyJapan          = 65
    OneCard                = 66
    PSE                    = 67
    Exito                  = 68
    Efecty                 = 69
    Paloto                 = 70
    PinValidda             = 71
    MangirKart             = 72
    BancoCreditoDePeru     = 73
    BBVAContinental        = 74
    SafetyPay              = 75
    PagoEfectivo           = 76
    Trustly                = 77
    UnionPay               = 78
    BitCoin                = 79
    Wallet                 = 128
    Valve                  = 129
    MasterComp             = 130
    Promotional            = 131
    MasterSubscription     = 134
    Payco                  = 135
    MobileWalletJapan      = 136
    OEMTicket              = 256
    Split                  = 512
    Complimentary          = 1024


class PackageStatus(IntEnum):
    Available   = 0
    Preorder    = 1
    Unavailable = 2
    Invalid     = 3


class PublishedFileRevision(IntEnum):
    Default               = 0
    Latest                = 1
    ApprovedSnapshot      = 2
    ApprovedSnapshotChina = 3
    RejectedSnapshot      = 4
    RejectedSnapshotChina = 5


class PublishedFileQueryType:
    RankedByVote                                  = 0
    RankedByPublicationDate                       = 1
    AcceptedForGameRankedByAcceptanceDate         = 2
    RankedByTrend                                 = 3
    FavouritedByFriendsRankedByPublicationDate    = 4
    CreatedByFriendsRankedByPublicationDate       = 5
    RankedByNumTimesReported                      = 6
    CreatedByFollowedUsersRankedByPublicationDate = 7
    NotYetRated                                   = 8
    RankedByTotalUniqueSubscriptions              = 9
    RankedByTotalVotesAsc                         = 10
    RankedByVotesUp                               = 11
    RankedByTextSearch                            = 12
    RankedByPlaytimeTrend                         = 13
    RankedByTotalPlaytime                         = 14
    RankedByAveragePlaytimeTrend                  = 15
    RankedByLifetimeAveragePlaytime               = 16
    RankedByPlaytimeSessionsTrend                 = 17
    RankedByLifetimePlaytimeSessions              = 18
    RankedByInappropriateContentRating            = 19
    RankedByBanContentCheck                       = 20
    RankedByLastUpdatedDate                       = 21


class PublishedFileType(IntEnum):
    Community              = 0   #: Normal Workshop item that can be subscribed to
    Microtransaction       = 1   #: Workshop item that is meant to be voted on for the purpose of selling in-game
    Collection             = 2   #: A collection of Workshop or Greenlight items
    Art                    = 3   #: Artwork
    Video                  = 4   #: External video
    Screenshot             = 5   #: Screenshot
    Game                   = 6   #: Greenlight game entry
    Software               = 7   #: Greenlight software entry
    Concept                = 8   #: Greenlight concept
    WebGuide               = 9   #: Steam web guide
    IntegratedGuide        = 10  #: Application integrated guide
    Merch                  = 11  #: Workshop merchandise meant to be voted on for the purpose of being sold
    ControllerBinding      = 12  #: Steam Controller bindings
    SteamworksAccessInvite = 13  #: Internal
    SteamVideo             = 14  #: Steam video
    GameManagedItem        = 15  #: Managed completely by the game, not the user, and not shown on the web
    Max                    = 16


class PublishedFileVisibility(IntEnum):
    Public      = 0
    FriendsOnly = 1
    Private     = 2
    Unlisted    = 3


class PublishedFileQueryFileType(IntEnum):
    Items                   = 0   #: Items.
    Collections             = 1   #: A collection of Workshop items.
    Art                     = 2   #: Artwork.
    Videos                  = 3   #: Videos.
    Screenshots             = 4   #: Screenshots.
    CollectionEligible      = 5   #: Items that can be put inside a collection.
    Games                   = 6   #: Unused.
    Software                = 7   #: Unused.
    Concepts                = 8   #: Unused.
    GreenlightItems         = 9   #: Unused.
    AllGuides               = 10  #: Guides.
    WebGuides               = 11  #: Steam web guide.
    IntegratedGuides        = 12  #: Application integrated guide.
    UsableInGame            = 13  #:
    Merch                   = 14  #: Workshop merchandise meant to be voted on for the purpose of being sold.
    ControllerBindings      = 15  #: Steam Controller bindings.
    SteamworksAccessInvites = 16  #: Used internally.
    ForSaleInGame           = 17  #: Workshop items that can be sold in-game.
    ReadyToUseItems         = 18  #: Workshop items that can be used right away by the user.
    WorkshopShowcases       = 19  #:
    GameManagedItems        = 20  #: Managed completely by the game, not the user, and not shown on the web.
# fmt: on


if __debug__:
    _ENUM_NAMES = {enum.__name__ for enum in Enum.__subclasses__() + IntEnum.__subclasses__()}
    assert _ENUM_NAMES.issubset(__all__), f"__all__ is not complete, missing {_ENUM_NAMES - set(__all__)}"


# shim for old enum names
def __getattr__(name: str) -> Any:
    if name == "ClanEvent":
        return EventType
    if name[0] == "E" and name[1:] in __all__:
        import warnings

        warnings.warn('Enums with "E" prefix are depreciated and scheduled for removal in V.1', DeprecationWarning)
        return globals()[name[1:]]

    raise AttributeError(name)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import sys
from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from .enums import UserBadge
from .game import StatefulGame
from .utils import DateTime

if TYPE_CHECKING:
    from .state import ConnectionState


__all__ = (
    "Badge",
    "FavouriteBadge",
    "UserBadges",
)


class BaseBadge:
    """
    Attributes
    ----------
    id
        The badge's ID.
    level
        The badge's level.
    game
        The game associated with the badge.
    """

    __slots__ = ()
    id: int
    level: int
    game: StatefulGame | None

    def __repr__(self) -> str:
        attrs = ("id", "level", "game")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.id == other.id


@dataclass(repr=False)
class FavouriteBadge(BaseBadge):
    """Represents a User's favourite badge.

    Attributes
    ----------
    id
        The badge's ID.
    level
        The badge's level.
    game
        The game associated with the badge.
    item_id
        The badge's item's ID.
    type
        The badge's type.
    border_colour
        The colour of the boarder of the badge.
    """

    __slots__ = ("id", "level", "game", "item_id", "type", "border_colour")

    id: int
    level: int
    game: StatefulGame | None
    item_id: int
    type: int
    border_colour: int


class Badge(BaseBadge):
    """Represents a Steam badge.

    Attributes
    ----------
    id
        The badge's ID.
    level
        The badge's level.
    game
        The game associated with the badge.
    xp
        The badge's XP.
    completion_time
        The time the badge was completed at.
    scarcity
        The scarcity of the badge.
    """

    __slots__ = ("id", "xp", "game", "level", "scarcity", "completion_time")

    def __init__(self, state: ConnectionState, data: dict[str, Any]):
        self.id = UserBadge.try_value(data["badgeid"])
        self.level: int = data["level"]
        self.xp: int = data["xp"]
        self.completion_time = DateTime.from_timestamp(data["completion_time"])
        self.scarcity: int = data["scarcity"]
        self.game = StatefulGame(state, id=data["appid"]) if "appid" in data else None


if TYPE_CHECKING or sys.version_info >= (3, 9):
    UserBadgeBase = Sequence[Badge]
else:
    UserBadgeBase = Sequence


class UserBadges(UserBadgeBase):
    """Represents a Steam :class:`~steam.User`'s badges/level.

    Attributes
    ----------
    level
        The badge's level.
    xp
        The badge's XP.
    xp_needed_to_level_up
        The amount of XP the user needs to level up.
    xp_needed_for_current_level
        The amount of XP the user's current level requires to achieve.
    badges
        A list of the user's badges.
    """

    __slots__ = (
        "xp",
        "level",
        "badges",
        "xp_needed_to_level_up",
        "xp_needed_for_current_level",
    )

    def __init__(self, state: ConnectionState, data: dict[str, Any]):
        self.level: int = data["player_level"]
        self.xp: int = data["player_xp"]
        self.xp_needed_to_level_up: int = data["player_xp_needed_to_level_up"]
        self.xp_needed_for_current_level: int = data["player_xp_needed_current_level"]
        self.badges: Sequence[Badge] = [Badge(state, data) for data in data["badges"]]

    def __repr__(self) -> str:
        attrs = ("level", "xp")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f'<UserBadges {" ".join(resolved)}>'

    def __len__(self) -> int:
        return len(self.badges)

    if not TYPE_CHECKING:

        def __getitem__(self, idx: Any) -> Any:
            return self.badges[idx]



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import Any

from typing_extensions import Self

from ..utils import StructIO
from .emsg import EMsg
from .protobufs import PROTOBUFS


class StructMessageMeta(type):
    def __new__(mcs, name: str, bases: tuple[type, ...], attrs: dict[str, Any]) -> Self:
        attrs["__slots__"] = slots = tuple(attrs.get("__annotations__", ()))
        exec(
            f"def __init__(self, {', '.join(f'{slot}=None' for slot in slots)}): "
            f"{' '.join(f'self.{slot} = {slot};' for slot in slots) or 'pass'}\n",
            {},
            attrs,
        )

        cls = super().__new__(mcs, name, bases, attrs)
        if cls.__module__ == __name__ and name != "StructMessage":
            PROTOBUFS[EMsg[name]] = cls  # type: ignore

        return cls


class StructMessage(metaclass=StructMessageMeta):
    def from_dict(self, dict: dict[str, Any]) -> Self:
        self.__init__(**dict)
        return self

    def to_dict(self, *_) -> dict[str, Any]:
        try:
            annotations = self.__annotations__
        except AttributeError:
            annotations = ()
        return {key: getattr(self, key) for key in annotations}

    def __bytes__(self) -> bytes:
        try:
            annotations = self.__annotations__
        except AttributeError:
            annotations = {}
        with StructIO() as io:
            for key, annotation in annotations.items():
                if annotation == "int":
                    io.write_u64(getattr(self, key))
                elif annotation == "bool":
                    io.write_u8(getattr(self, key))

            return io.buffer

    def parse(self, data: bytes) -> Self:
        raise NotImplementedError


class ClientGetFriendsWhoPlayGame(StructMessage):
    app_id: int


class ClientGetFriendsWhoPlayGameResponse(StructMessage):
    eresult: int
    app_id: int
    friends: list[int]

    def parse(self, data: bytes) -> Self:
        with StructIO(data) as io:
            self.eresult = io.read_u32()
            self.app_id = io.read_u64()
            self.friends = [io.read_u64() for _ in range(io.read_u32())]

        return self



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import re
from dataclasses import dataclass
from datetime import timedelta
from typing import TYPE_CHECKING, Any, Sequence

from .enums import LicenseFlag, LicenseType, PaymentMethod
from .game import PartialGamePriceOverview, StatefulGame
from .utils import DateTime, Intable

if TYPE_CHECKING:
    from .manifest import Depot, GameInfo, HeadlessDepot, PackageInfo
    from .message import Authors
    from .protobufs.client_server import CMsgClientLicenseListLicense
    from .protobufs.store import PurchaseReceiptInfo
    from .state import ConnectionState
    from .types import game, package


__all__ = (
    "Package",
    "FetchedPackage",
    "FetchedGamePackage",
    "License",
)


class Package:
    """Represents information on a package which is a collection of one or more applications and depots.

    Read more on `steamworks <https://partner.steamgames.com/doc/store/application/packages>`_.
    """

    __slots__ = (
        "id",
        "name",
    )

    def __init__(self, id: Intable, name: str | None = None):
        self.id = int(id)
        """The package's ID."""
        self.name = name
        """The package's name."""

    def __eq__(self, other: object) -> bool:
        return self.id == other.id if isinstance(other, Package) else NotImplemented

    def __hash__(self) -> int:
        return hash(self.id)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id}, name={self.name!r})"


class StatefulPackage(Package):
    """A package with state."""

    __slots__ = ("_state",)

    def __init__(self, state: ConnectionState, **kwargs: Any):
        super().__init__(**kwargs)
        self._state = state

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} id={self.id}>"

    async def games(self) -> list[StatefulGame]:
        """Fetches this package's games."""
        fetched = await self.fetch()
        return fetched._games

    async def fetch(self) -> FetchedPackage:
        """Fetches this package's information. Shorthand for:

        .. code-block:: python3

            package = await client.fetch_package(package)
        """
        package = await self._state.client.fetch_package(self.id)
        if package is None:
            raise ValueError("Fetched package was not valid.")
        return package

    async def info(self) -> PackageInfo:
        """Shorthand for:

        .. code-block:: python3

            (info,) = await client.fetch_product_info(packages=[package])
        """
        _, (info,) = await self._state.fetch_product_info(package_ids=(self.id,))
        return info

    async def games_info(self) -> list[GameInfo]:
        """Shorthand for:

        .. code-block:: python3

            infos = await client.fetch_product_info(games=await package.games())
        """
        games = await self.games()
        if not games:
            return []
        infos, _ = await self._state.fetch_product_info(game.id for game in games)
        return infos

    async def depots(self) -> Sequence[Depot | HeadlessDepot]:
        """Fetches this package's depots."""
        try:
            depot_ids = self.depot_ids  # type: ignore
        except AttributeError:
            info = await self.info()
            depot_ids = info.depot_ids

        games_info = await self.games_info()

        return [
            depot
            for game_info in games_info
            for branch in game_info._branches.values()
            for depot in branch.depots
            if depot.id in depot_ids
        ] + [
            depot for game_info in games_info for depot in game_info.headless_depots if depot.id in depot_ids
        ]  # type: ignore

    # TODO .manifests, fetch_manifest


@dataclass
class PackagePriceOverview(PartialGamePriceOverview):
    individual: int


class FetchedPackage(StatefulPackage):
    name: str

    def __init__(self, state: ConnectionState, data: package.FetchedPackage):
        super().__init__(state, name=data["name"], id=data["packageid"])
        self._games = [StatefulGame(state, id=app["id"], name=app["name"]) for app in data["apps"]]
        self.description = data["page_content"]
        self.created_at = DateTime.parse_steam_date(data["release_date"]["date"], full_month=False)
        self.logo_url = data["small_logo"]
        self.logo_url = data["header_image"]
        platforms = data["platforms"]
        self._on_windows = platforms["windows"]
        self._on_mac_os = platforms["mac"]
        self._on_linux = platforms["linux"]
        self.price_overview = PackagePriceOverview(**data["price"])

    def is_free(self) -> bool:
        """Whether the game is free."""
        return self.price_overview.final == 0

    def is_on_windows(self) -> bool:
        """Whether the game is playable on Windows."""
        return self._on_windows

    def is_on_mac_os(self) -> bool:
        """Whether the game is playable on macOS."""
        return self._on_mac_os

    def is_on_linux(self) -> bool:
        """Whether the game is playable on Linux."""
        return self._on_linux


@dataclass
class FetchedGamePackagePriceOverview:
    percent_discount: int
    final: int


class FetchedGamePackage(StatefulPackage):
    __slots__ = (
        "_is_free",
        "price_overview",
    )

    def __init__(self, state: ConnectionState, data: game.PackageGroupSub):
        name, _, _ = data["option_text"].rpartition(" - ")
        super().__init__(state, name=name, id=data["packageid"])
        self._is_free = data["is_free_license"]
        self.price_overview = FetchedGamePackagePriceOverview(
            int(re.search(r"-?(\d)", data["percent_savings_text"])[0]),
            data["price_in_cents_with_discount"],  # this isn't always in cents
        )

    def is_free(self) -> bool:
        return self._is_free


class License(StatefulPackage):
    """Represents a License to a package the client user has access to."""

    __slots__ = (
        "owner",
        "type",
        "flags",
        "created_at",
        "master_package",
        "next_process_at",
        "time_limit",
        "time_used",
        "payment_method",
        "purchase_country_code",
        "territory_code",
        "change_number",
        "initial_period",
        "initial_time_unit",
        "renewal_period",
        "renewal_time_unit",
        "access_token",
    )

    def __init__(self, state: ConnectionState, proto: CMsgClientLicenseListLicense, owner: Authors):
        super().__init__(state, id=proto.package_id)
        self.owner: Authors = owner
        """The license's owner."""
        self.type = LicenseType.try_value(proto.license_type)
        """The license's type."""
        self.flags = LicenseFlag.try_value(proto.flags)
        """The license's flags."""
        self.created_at = DateTime.from_timestamp(proto.time_created) if proto.time_created else None
        """The license's creation date."""
        self.master_package = StatefulPackage(state, id=proto.master_package_id) if proto.master_package_id else None
        """The license's master package."""
        self.next_process_at = DateTime.from_timestamp(proto.time_next_process)
        """The date when the license will be processed."""
        self.time_limit = timedelta(minutes=proto.minute_limit) if proto.minute_limit else None
        """The time limit for the license."""
        self.time_used = timedelta(minutes=proto.minutes_used)
        """The time the license has been used."""
        self.payment_method = PaymentMethod.try_value(proto.payment_method)
        """The payment method used for the license."""
        self.purchase_country_code = proto.purchase_country_code
        """The country code of the license's purchase."""
        self.territory_code = proto.territory_code
        """The license's territory code."""
        self.change_number = proto.change_number
        """The license's current change number."""
        self.initial_period = proto.initial_period
        """The license's initial period."""
        self.initial_time_unit = proto.initial_time_unit
        """The license's initial time unit."""
        self.renewal_period = timedelta(minutes=proto.renewal_period)
        """The license's renewal period."""
        self.renewal_time_unit = proto.renewal_time_unit
        """The license's renewal time unit."""
        self.access_token = proto.access_token
        """The license's access token."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} owner={self.owner!r} flags={self.flags!r}>"

    @property
    def time_remaining(self) -> timedelta | None:
        """The amount of time that this license can be used for."""
        if self.flags & LicenseFlag.Expired:
            return
        if self.time_limit is None:
            return
        return self.time_limit - self.time_used


# class Transaction:
#     def __init__(self, state: ConnectionState, proto: PurchaseReceiptInfo):
#         self.id = proto.transactionid
#         self.items = [
#             TransactionItem(
#                 Package(state, item.packageid), StatefulGame(state, id=item.appid), item.line_item_description
#             )
#             for item in proto.line_items
#         ]
#
#         self.package = Package(state, proto.packageid)
#         self.purchase_status = proto.purchase_status
#         self.created_at = DateTime.from_timestamp(proto.transaction_time)
#         self.payment_method = PaymentMethod.try_value(proto.payment_method)
#         self.base_price = proto.base_price
#         self.total_discount = proto.total_discount
#         self.tax = proto.tax
#         self.shipping = proto.shipping
#         self.currency_code = proto.currency_code
#         self.country_code = proto.country_code
#
#
# @dataclass
# class TransactionItem:
#     package_id: Package
#     game: StatefulGame
#     description: str



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .abc import SteamID
    from .clan import Clan
    from .enums import FriendRelationship
    from .state import ConnectionState
    from .user import User


__all__ = (
    "Invite",
    "UserInvite",
    "ClanInvite",
)


class Invite:
    __slots__ = ("invitee", "relationship", "_state")

    def __init__(self, state: ConnectionState, invitee: User | SteamID, relationship: FriendRelationship | None):
        self._state = state
        self.invitee = invitee
        self.relationship = relationship

    async def accept(self) -> None:
        raise NotImplementedError()

    async def decline(self) -> None:
        raise NotImplementedError()


class UserInvite(Invite):
    """Represents a invite from a user to become their friend.

    Attributes
    ----------
    invitee
        The user who sent the invite.
    relationship
        The relationship you have with the invitee.
    """

    __slots__ = ()

    def __repr__(self) -> str:
        attrs = ("invitee",)
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<UserInvite {' '.join(resolved)}>"

    async def accept(self) -> None:
        """Accept the invite request."""
        await self._state.http.accept_user_invite(self.invitee.id64)

    async def decline(self) -> None:
        """Decline the invite request."""
        await self._state.http.decline_user_invite(self.invitee.id64)


class GroupInvite(Invite):
    ...


class ClanInvite(Invite):
    """Represents a invite to join a :class:`~steam.Clan` from a user.

    Attributes
    ----------
    clan
        The clan to join.
    invitee
        The user who sent the invite.
    relationship
        The relationship you have with the clan.
    """

    __slots__ = ("clan",)

    def __init__(
        self,
        state: ConnectionState,
        invitee: User | SteamID,
        clan: Clan | SteamID,
        relationship: FriendRelationship | None,
    ):
        super().__init__(state, invitee, relationship)
        self.clan = clan

    def __repr__(self) -> str:
        attrs = ("invitee", "clan")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<ClanInvite {' '.join(resolved)}>"

    async def accept(self) -> None:
        """Accept the invite request."""
        await self._state.respond_to_clan_invite(self.clan.id64, True)

    async def decline(self) -> None:
        """Decline the invite request."""
        await self._state.respond_to_clan_invite(self.clan.id64, False)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from typing_extensions import Self

if TYPE_CHECKING:
    from .chat import Member
    from .clan import Clan
    from .group import Group
    from .protobufs import chat
    from .state import ConnectionState

__all__ = (
    "Role",
    "RolePermissions",
)


class Role:
    __slots__ = ("id", "name", "ordinal", "clan", "group", "permissions", "_state")

    def __init__(self, state: ConnectionState, group: Clan | Group, role: chat.Role, permissions: chat.RoleActions):
        self._state = state
        self.id = int(role.role_id)
        self.name = role.name[len("#ChatRoomRole_") :]  # str.removeprefix
        self.ordinal = role.ordinal

        from .clan import Clan

        if isinstance(group, Clan):
            self.clan = group
            self.group = None
        else:
            self.group = group
            self.clan = None
        self.permissions = RolePermissions(permissions)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} id={self.id}>"

    @property
    def members(self) -> list[Member]:
        chat_group = self.clan or self.group
        assert chat_group is not None
        return [m for m in chat_group.members if self in m.roles]

    async def edit(self, *, name: str | None = None, permissions: RolePermissions | None = None) -> None:
        chat_group = self.group or self.clan
        assert chat_group is not None
        chat_group_id = chat_group._id
        if name is not None:
            await self._state.edit_role_name(self.id, chat_group_id, name=name)
        if permissions is not None:
            await self._state.edit_role_permissions(self.id, chat_group_id, permissions=permissions)

    async def delete(self) -> None:
        chat_group = self.group or self.clan
        assert chat_group is not None
        chat_group_id = chat_group._id
        await self._state.delete_role(self.id, chat_group_id)


class RolePermissions:
    __slots__ = (
        "kick",
        "ban_members",
        "invite",
        "manage_group",
        "send_messages",
        "read_message_history",
        "change_group_roles",
        "change_user_roles",
        "mention_all",
        "set_watching_broadcast",
    )

    def __init__(self, proto: chat.RoleActions):
        self.kick = proto.can_kick
        self.ban_members = proto.can_ban
        self.invite = proto.can_invite
        self.manage_group = proto.can_change_tagline_avatar_name
        self.send_messages = proto.can_chat
        self.read_message_history = proto.can_view_history
        self.change_group_roles = proto.can_change_group_roles
        self.change_user_roles = proto.can_change_user_roles
        self.mention_all = proto.can_mention_all
        self.set_watching_broadcast = proto.can_set_watching_broadcast

    def copy(self) -> Self:
        return self.__class__(chat.RoleActions(**self.to_dict()))

    __copy__ = copy

    def to_dict(self) -> dict[str, bool]:
        return {
            "can_kick": self.kick,
            "can_ban": self.ban_members,
            "can_invite": self.invite,
            "can_change_tagline_avatar_name": self.manage_group,
            "can_chat": self.send_messages,
            "can_view_history": self.read_message_history,
            "can_change_group_roles": self.change_group_roles,
            "can_change_user_roles": self.change_user_roles,
            "can_mention_all": self.mention_all,
            "can_set_watching_broadcast": self.set_watching_broadcast,
        }



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
import re
import warnings
from collections.abc import Sequence
from datetime import datetime
from typing import TYPE_CHECKING, TypeVar, overload

from bs4 import BeautifulSoup
from typing_extensions import Literal, Self

from . import utils
from ._const import HTML_PARSER
from .abc import Commentable, SteamID, _CommentableKwargs
from .channel import ClanChannel
from .chat import ChatGroup, Member
from .enums import EventType, Language, Type
from .errors import HTTPException
from .event import Announcement, Event
from .game import Game, StatefulGame
from .iterators import AnnouncementsIterator, EventIterator
from .protobufs import chat
from .utils import DateTime

if TYPE_CHECKING:
    from .state import ConnectionState
    from .types.id import ID32
    from .user import User

__all__ = ("Clan", "ClanMember")

BoringEventT = TypeVar(
    "BoringEventT",
    Literal[EventType.Other],
    Literal[EventType.Chat],
    Literal[EventType.Party],
    Literal[EventType.Meeting],
    Literal[EventType.SpecialCause],
    Literal[EventType.MusicAndArts],
    Literal[EventType.Sports],
    Literal[EventType.Trip],
    covariant=True,
)


class ClanMember(Member):
    group: None
    clan: Clan

    def __init__(self, state: ConnectionState, clan: Clan, user: User, proto: chat.Member):
        super().__init__(state, clan, user, proto)
        self.clan = clan


class Clan(ChatGroup[ClanMember, ClanChannel], Commentable, utils.AsyncInit):
    """Represents a Steam clan.

    .. container:: operations

        .. describe:: x == y

            Checks if two clans are equal.

        .. describe:: str(x)

            Returns the clan's name.

    Attributes
    ------------
    name
        The name of the clan.
    avatar_url
        The icon url of the clan. Uses the large (184x184 px) image url.
    content
        The content of the clan.
    tagline
        The clan's tagline.
    member_count
        The amount of users in the clan.
    online_count
        The amount of users currently online.
    active_member_count
        The amount of currently users in the clan's chat room.
    in_game_count
        The amount of user's currently in game.
    created_at
        The time the clan was created_at.
    language
        The language set for the clan.
    location
        The location set for the clan.
    game
        The clan's associated game.
    owner
        The clan's owner.
    admins
        A list of the clan's administrators.
    mods
        A list of the clan's moderators.
    """

    __slots__ = (
        "content",
        "created_at",
        "language",
        "location",
        "member_count",
        "in_game_count",
        "online_count",
        "admins",
        "mods",
        "community_url",
        "is_game_clan",
    )

    # TODO more to implement https://github.com/DoctorMcKay/node-steamcommunity/blob/master/components/groups.js
    # Clan.requesting_membership
    # Clan.respond_to_requesting_membership(*users, approve)
    # Clan.respond_to_all_requesting_membership(approve)

    # V1
    # Clan.headline
    # Clan.summary
    # Clan.flags https://cs.github.com/SteamDatabase/SteamTracking/blob/5c4420496f18384bea932f4535ee1a87fd9271e4/Structs/enums.steamd#L526
    # or more likely https://cs.github.com/SteamDatabase/SteamTracking/blob/5c4420496f18384bea932f4535ee1a87fd9271e4/Structs/enums.steamd#L3177

    content: str
    created_at: datetime | None
    member_count: int
    online_count: int
    in_game_count: int
    language: Language
    location: str
    mods: list[ClanMember]
    admins: list[ClanMember]
    is_game_clan: bool

    def __init__(self, state: ConnectionState, id: int):
        super().__init__(id, type=Type.Clan)
        self._state = state

    async def __ainit__(self) -> None:
        resp = await self._state.http._session.get(super().community_url)  # type: ignore
        text = await resp.text()  # technically we loose proper request handling here
        if not self.id64:
            search = utils.CLAN_ID64_FROM_URL_REGEX.search(text)
            if search is None:
                raise ValueError("unreachable code reached")
            super().__init__(search["steamid"], type=Type.Clan)

        soup = BeautifulSoup(text, HTML_PARSER)
        _, __, name = soup.title.text.rpartition(" :: ")
        self.name = name
        content = soup.find("meta", property="og:description")
        self.content = content["content"] if content is not None else None
        icon_url = soup.find("link", rel="image_src")
        self.avatar_url = icon_url["href"] if icon_url else None
        self.is_game_clan = "games" in resp.url.parts
        self.community_url = str(resp.url)
        if self.is_game_clan:
            for entry in soup.find_all("div", class_="actionItem"):
                a = entry.a
                if a is not None:
                    href = a.get("href", "")
                    match = re.findall(r"store.steampowered.com/app/(\d+)", href)
                    if match:
                        self.game = StatefulGame(self._state, id=match[0])
        stats = soup.find("div", class_="grouppage_resp_stats")
        if stats is None:
            return

        for stat in stats.find_all("div", class_="groupstat"):
            if "Founded" in stat.text:
                text = stat.text.split("Founded")[1].strip()
                if ", " not in stat.text:
                    text = f"{text}, {DateTime.now().year}"
                self.created_at = utils.DateTime.parse_steam_date(text)
            if "Language" in stat.text:
                self.language = stat.text.split("Language")[1].strip()
            if "Location" in stat.text:
                self.location = stat.text.split("Location")[1].strip()

        for count in stats.find_all("div", class_="membercount"):
            if "MEMBERS" in count.text:
                self.member_count = int(count.text.split("MEMBERS")[0].strip().replace(",", ""))
            if "IN-GAME" in count.text:
                self.in_game_count = int(count.text.split("IN-GAME")[0].strip().replace(",", ""))
            if "ONLINE" in count.text:
                self.online_count = int(count.text.split("ONLINE")[0].strip().replace(",", ""))

        admins = []
        mods = []
        is_admins = None
        for fields in soup.find_all("div", class_="membergrid"):
            for field in fields.find_all("div"):
                if "Administrators" in field.text:
                    is_admins = True
                    continue
                if "Moderators" in field.text:
                    is_admins = False
                    continue
                if "Members" in field.text:
                    break

                try:
                    account_id = int(field["data-miniprofile"])
                except KeyError:
                    continue
                else:
                    if is_admins is None:
                        continue
                    if is_admins:
                        admins.append(account_id)
                    else:
                        mods.append(account_id)

        users = await self._state.client.fetch_users(*admins, *mods)
        self.admins = [user for user in users if user and user.id in admins]
        self.mods = [user for user in users if user and user.id in mods]

    @classmethod
    async def _from_proto(
        cls,
        state: ConnectionState,
        proto: chat.GetChatRoomGroupSummaryResponse,
        *,
        maybe_chunk: bool = True,
    ) -> Self:
        self = await super()._from_proto(state, proto, id=proto.clanid, maybe_chunk=maybe_chunk)
        return await self

    # TODO properties for admins and mods when chunked?

    async def chunk(self) -> Sequence[ClanMember]:
        self._members = dict.fromkeys(self._partial_members)  # type: ignore
        if len(self._partial_members) <= 100:
            # TODO might be if self.flags & ClanFlags.Large (2)?
            for id, member in self._partial_members.items():
                user = self._state.get_user(id)
                if user is None:
                    await asyncio.sleep(0)
                    user = await self._state._maybe_user(utils.make_id64(id))  # TODO maybe users
                member = ClanMember(self._state, self, user, member)
                self._members[member.id] = member
            return await super().chunk()

        # these actually need fetching
        view_id = self._state.chat_group_to_view_id[self._id]
        users: dict[ID32, User] = {
            user.id: user
            for users in await asyncio.gather(
                *(
                    self._state.request_chat_group_members(
                        self._id,
                        view_id,
                        client_change_number + 1,  # steam doesn't send responses if they're 0
                        start + 1,
                        stop,
                    )
                    for client_change_number, (start, stop) in enumerate(
                        utils._int_chunks(len(self._partial_members), 100)
                    )
                )
            )
            for user in users
        }
        for id, member in self._partial_members.items():
            try:
                user = users[id]
            except KeyError:
                # steam doesn't include the first user cause ???, this however, isn't that big a deal.
                user = await self._state._maybe_user(utils.make_id64(id))
                if isinstance(user, SteamID):
                    continue
            member = ClanMember(self._state, self, users[id], member)
            self._members[member.id] = member

        return await super().chunk()

    @property
    def _commentable_kwargs(self) -> _CommentableKwargs:
        return {
            "id64": self.id64,
            "thread_type": 12,
        }

    async def fetch_members(self) -> list[SteamID]:
        """Fetches a clan's member list.

        Note
        ----
        This can be a very slow operation due to the rate limits on this endpoint.
        """

        async def getter(i: int) -> BeautifulSoup:
            try:
                resp = await self._state.http.get(
                    f"{self.community_url}/members", params={"p": i + 1, "content_only": "true"}
                )
            except HTTPException:
                await asyncio.sleep(20)
                return await getter(i)
            else:
                soup = BeautifulSoup(resp, HTML_PARSER)
                for s in soup.find_all("div", id="memberList"):
                    for user in s.find_all("div", class_="member_block"):
                        ret.append(SteamID(user["data-miniprofile"]))

                return soup

        ret: list[SteamID] = []
        soup = await getter(0)
        number_of_pages = int(re.findall(r"\d* - (\d*)", soup.find("div", class_="group_paging").text)[0])
        await asyncio.gather(*(getter(i) for i in range(1, number_of_pages)))
        return ret

    @property
    def description(self) -> str:
        """An alias to :attr:`content`.

        .. deprecated:: 0.8.0

            Use :attr:`content` instead.
        """
        warnings.warn("Clan.description is deprecated, use Clan.content instead", DeprecationWarning, stacklevel=2)
        return self.content

    @property
    def icon_url(self) -> str:
        """An alias to :attr:`avatar_url`.

        .. deprecated:: 0.8.0

            Use :attr:`avatar_url` instead.
        """
        warnings.warn("Clan.icon_url is deprecated, use Clan.avatar_url instead", DeprecationWarning, stacklevel=2)
        return self.avatar_url

    async def join(self, *, invite_code: str | None = None) -> None:
        """Joins the clan."""
        await self._state.http.join_clan(self.id64)
        await super().join(invite_code=invite_code)

    async def leave(self) -> None:
        """Leaves the clan."""
        await super().leave()
        await self._state.http.leave_clan(self.id64)

    # event/announcement stuff

    async def fetch_event(self, id: int) -> Event[EventType]:
        """Fetch an event from its ID.

        Parameters
        ----------
        id
            The ID of the event.
        """
        data = await self._state.http.get_clan_events(self.id, [id])
        events = data["events"]
        if not events:
            raise ValueError(f"Event {id} not found")
        return await Event(self._state, self, events[0])

    async def fetch_announcement(self, id: int) -> Announcement:
        """Fetch an announcement from its ID.

        Parameters
        ----------
        id
            The ID of the announcement.
        """
        data = await self._state.http.get_clan_announcement(self.id, id)
        announcement = data["events"][0]
        return await Announcement(self._state, self, announcement)

    def events(
        self,
        *,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> EventIterator:
        """An :class:`~steam.iterators.AsyncIterator` over a clan's :class:`steam.Event`\\s.

        Examples
        --------

        Usage:

        .. code-block:: python3

            async for event in clan.events(limit=10):
                print(event.author, "made an event", event.name, "starting at", event.starts_at)

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of events to search through. Default is ``100``. Setting this to ``None`` will fetch all
            the clan's events, but this will be a very slow operation.
        before
            A time to search for events before.
        after
            A time to search for events after.

        Yields
        ---------
        :class:`~steam.Event`
        """
        return EventIterator(self, self._state, limit, before, after)

    def announcements(
        self,
        *,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> AnnouncementsIterator:
        """An :class:`~steam.iterators.AsyncIterator` over a clan's :class:`steam.Announcement`\\s.

        Examples
        --------

        Usage:

        .. code-block:: python3

            async for announcement in clan.announcements(limit=10):
                print(
                    announcement.author,
                    "made an announcement",
                    announcement.name,
                    "at",
                    announcement.created_at,
                )

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of announcements to search through. Default is ``100``. Setting this to ``None`` will
            fetch all of the clan's announcements, but this will be a very slow operation.
        before
            A time to search for announcements before.
        after
            A time to search for announcements after.

        Yields
        ---------
        :class:`~steam.Announcement`
        """
        return AnnouncementsIterator(self, self._state, limit, before, after)

    @overload
    async def create_event(
        self,
        name: str,
        content: str,
        *,
        type: Literal[EventType.Game] = ...,
        starts_at: datetime | None = ...,
        game: Game,
        server_address: str | None = ...,
        server_password: str | None = ...,
    ) -> Event[Literal[EventType.Game]]:
        ...

    @overload
    async def create_event(
        self,
        name: str,
        content: str,
        *,
        type: BoringEventT = EventType.Other,
        starts_at: datetime | None = None,
    ) -> Event[BoringEventT]:
        ...

    async def create_event(
        self,
        name: str,
        content: str,
        *,
        type: Literal[
            EventType.Other,
            EventType.Chat,
            EventType.Game,
            # ClanEvent.Broadcast,  # TODO need to wait until implementing stream support for this
            EventType.Party,
            EventType.Meeting,
            EventType.SpecialCause,
            EventType.MusicAndArts,
            EventType.Sports,
            EventType.Trip,
        ] = EventType.Other,
        game: Game | None = None,
        starts_at: datetime | None = None,
        server_address: str | None = None,
        server_password: str | None = None,
    ) -> Event[EventType]:
        """Create an event.

        Parameters
        ----------
        name
            The name of the event
        content
            The content for the event.
        type
            The type of the event, defaults to :attr:`ClanEvent.Other`.
        game
            The game that will be played in the event. Required if type is :attr:`ClanEvent.Game`.
        starts_at
            The time the event will start at.
        server_address
            The address of the server that the event will be played on. This is only allowed if ``type`` is
            :attr:`ClanEvent.Game`.
        server_password
            The password for the server that the event will be played on. This is only allowed if ``type`` is
            :attr:`ClanEvent.Game`.

        Note
        ----
        It is recommended to use a timezone aware datetime for ``start``.

        Returns
        -------
        The created event.
        """

        resp = await self._state.http.create_clan_event(
            self.id64,
            name,
            content,
            f"{type.name}Event",
            str(game.id) if game is not None else "",
            server_address or "",
            server_password or "",
            starts_at,
        )
        soup = BeautifulSoup(resp, HTML_PARSER)
        for element in soup.find_all("div", class_="eventBlockTitle"):
            a = element.a
            if a is not None and a.text == name:  # this is bad?
                _, __, id = a["href"].rpartition("/")
                event = await self.fetch_event(int(id))
                self._state.dispatch("event_create", event)
                return event
        raise ValueError

    async def create_announcement(
        self,
        name: str,
        content: str,
        hidden: bool = False,
    ) -> Announcement:
        """Create an announcement.

        Parameters
        ----------
        name
            The name of the announcement.
        content
            The content of the announcement.
        hidden
            Whether the announcement should initially be hidden.

        Returns
        -------
        The created announcement.
        """
        await self._state.http.create_clan_announcement(self.id64, name, content, hidden)
        resp = await self._state.http.get(f"{self.community_url}/announcements", params={"content_only": "true"})
        soup = BeautifulSoup(resp, HTML_PARSER)
        for element in soup.find_all("div", class_="announcement"):
            a = element.a
            if a is not None and a.text == name:  # this is bad?
                _, __, id = a["href"].rpartition("/")
                announcement = await self.fetch_announcement(int(id))
                self._state.dispatch("announcement_create", announcement)
                return announcement

        raise ValueError



"""
steam.protobufs
~~~~~~~~~~~~~~~
The protobufs for the Steam API.

Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE

Contains large portions of:
https://github.com/ValvePython/steam/tree/master/steam/core/msg/__init__.py
The appropriate license is in LICENSE
"""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, Generic, TypeVar

import betterproto
from typing_extensions import TypeAlias

from ..enums import IntEnum, Result
from .emsg import *
from .headers import *
from .protobufs import *
from .struct_messages import StructMessage
from .unified import *

M = TypeVar("M", bound="betterproto.Message | StructMessage")
GetProtoType: TypeAlias = "type[betterproto.Message] | None"
betterproto.safe_snake_case = do_nothing_case


def get_cmsg(msg: IntEnum) -> GetProtoType:
    """Get a protobuf from its EMsg."""
    try:
        return PROTOBUFS[msg]
    except KeyError:
        return None


def get_um(name: str, request: bool = True) -> GetProtoType:
    """Get the protobuf for a certain Unified Message."""
    try:
        return UMS[f"{name}#1_{'Request' if request else 'Response'}"]
    except KeyError:
        return None


class MsgBase(Generic[M]):
    __slots__ = ("header", "body", "payload", "skip")
    header: ExtendedMsgHdr | MsgHdr | MsgHdrProto | GCMsgHdr | GCMsgHdrProto

    if sys.version_info < (3, 9):  # see https://github.com/python/cpython/issues/83349 for the rationale behind this

        def __new__(cls, *args: Any, **kwargs: Any):
            return object.__new__(cls)

    def __init__(self, msg: IntEnum, data: bytes | None, **kwargs: Any):
        self.msg = msg
        self.body: M = None  # type: ignore
        self.payload: bytes | None = data[self.skip :] if data else None

        self.parse()
        if kwargs and self.body is not None:
            self.body.from_dict(kwargs)

    def __repr__(self) -> str:
        attrs = ("msg", "header")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        if self.body is not None:
            resolved.extend(f"{k}={v!r}" for k, v in self.body.to_dict(do_nothing_case).items())
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __bytes__(self) -> bytes:
        return bytes(self.header) + bytes(self.body)

    def _parse(self, proto: GetProtoType) -> None:
        if proto:
            self.body = proto()
            if self.payload:
                self.body.parse(self.payload)

    @property
    def msg(self) -> IntEnum:
        return self.header.msg

    @msg.setter
    def msg(self, value: IntEnum) -> None:
        self.header.msg = value

    @property
    def steam_id(self) -> int:
        return self.header.body.steam_id

    @steam_id.setter
    def steam_id(self, value: int) -> None:
        self.header.body.steam_id = value

    @property
    def session_id(self) -> int:
        return self.header.body.session_id

    @session_id.setter
    def session_id(self, value: int) -> None:
        self.header.body.session_id = value

    @property
    def result(self) -> Result:
        return Result.try_value(getattr(self.body, "eresult", 0) or self.header.body.eresult)

    def parse(self) -> None:
        proto = get_cmsg(self.msg)
        self._parse(proto)


if not TYPE_CHECKING:
    MsgBase.__class_getitem__ = classmethod(
        lambda cls, params: cls
    )  # really don't want this to be picked up by linters


class Msg(MsgBase[M]):
    """A wrapper around received messages."""

    __slots__ = ()

    def __init__(
        self,
        msg: EMsg,
        data: bytes | None = None,
        extended: bool = False,
        **kwargs: Any,
    ):
        self.header: ExtendedMsgHdr | MsgHdr = ExtendedMsgHdr(data) if extended else MsgHdr(data)
        self.skip = self.header.STRUCT.size
        super().__init__(msg, data, **kwargs)


class MsgProto(MsgBase[M]):
    """A wrapper around received protobuf messages."""

    __slots__ = ()

    def __init__(
        self,
        msg: EMsg,
        data: bytes | None = None,
        um_name: str | None = None,
        **kwargs: Any,
    ):
        self.header: MsgHdrProto = MsgHdrProto(data)
        if um_name:
            self.header.body.job_name_target = um_name
        self.skip = self.header.length
        super().__init__(msg, data, **kwargs)

    def parse(self) -> None:
        if self.msg in (
            EMsg.ServiceMethod,
            EMsg.ServiceMethodResponse,
            EMsg.ServiceMethodSendToClient,
            EMsg.ServiceMethodCallFromClient,
        ):
            name = self.header.body.job_name_target
            if name:
                name, _, _ = name.partition("#")
                self.header.body.job_name_target = name
                proto = get_um(name, self.msg in (EMsg.ServiceMethod, EMsg.ServiceMethodCallFromClient))
            else:
                proto = None

        else:
            proto = get_cmsg(self.msg)

        self._parse(proto)


class GCMsg(MsgBase[M]):
    """A wrapper around received GC messages, mainly for extensions."""

    __slots__ = ()

    def __init__(
        self,
        msg: IntEnum,
        data: bytes | None = None,
        **kwargs: Any,
    ):
        self.header: GCMsgHdr = GCMsgHdr(data)
        self.skip = self.header.STRUCT.size
        super().__init__(msg, data, **kwargs)


class GCMsgProto(MsgBase[M]):
    """A wrapper around received GC protobuf messages, mainly for extensions."""

    __slots__ = ()

    def __init__(
        self,
        msg: IntEnum,
        data: bytes | None = None,
        **kwargs: Any,
    ):
        self.header: GCMsgHdrProto = GCMsgHdrProto(data)
        self.skip = self.header.length
        super().__init__(msg, data, **kwargs)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from datetime import datetime
from io import BytesIO
from typing import TYPE_CHECKING, Sequence

from yarl import URL as URL_

from .abc import Awardable, Commentable, _CommentableKwargs
from .enums import Language, PublishedFileRevision, PublishedFileType, PublishedFileVisibility
from .game import StatefulGame
from .models import URL, _IOMixin
from .reaction import AwardReaction
from .utils import DateTime

if TYPE_CHECKING:
    from .manifest import Manifest
    from .message import Authors
    from .protobufs.published_file import PublishedFileDetails
    from .state import ConnectionState


__all__ = (
    "PublishedFile",
    "DetailsPreview",
    "PreviewInfo",
    "PublishedFileChild",
    "PublishedFileChange",
)


@dataclass
class DetailsPreview(_IOMixin):
    _state: ConnectionState
    id: int
    position: int
    url: str
    size: int
    filename: str
    type: int

    @property
    def ugc_id(self) -> int:
        return int(URL_(self.url).parts[2])


@dataclass
class PreviewInfo(_IOMixin):
    _state: ConnectionState
    ugc_id: int
    size: int
    url: str


@dataclass
class PublishedFileChild:
    _state: ConnectionState
    id: int
    position: int
    type: int
    parent: PublishedFile

    async def fetch(
        self,
        *,
        revision: PublishedFileRevision = PublishedFileRevision.Default,
        language: Language | None = None,
    ) -> PublishedFile:
        """Resolves this child into a full file.

        Parameters
        ----------
        revision
            The revision of the file to fetch.
        language
            The language to fetch the file in. If ``None``, the current language is used.
        """
        (file,) = await self._state.fetch_published_files((self.id,), revision, language)
        assert file
        return file

    async def manifest(self) -> Manifest:
        """Fetch the manifest for this child."""
        resolved = await self.fetch()
        return await resolved.manifest()


@dataclass
class PublishedFileChange:
    __slots__ = ("description", "created_at")
    description: str
    created_at: datetime


class PublishedFile(Commentable, Awardable):
    """Represents a published file on SteamPipe."""

    _AWARDABLE_TYPE = 2
    __slots__ = (
        "_state",
        "id",
        "name",
        "author",
        "game",
        "created_with",
        "created_at",
        "updated_at",
        "url",
        "manifest_id",
        "revision",
        "available_revisions",
        "change_number",
        "type",
        "views",
        "youtube_id",
        "description",
        "visibility",
        "flags",
        "accepted",
        "show_subscribe_all",
        "number_of_developer_comments",
        "number_of_public_comments",
        "spoiler",
        "children",
        "filename",
        "size",
        "cdn_url",
        "preview",
        "previews",
        "banned",
        "ban_reason",
        "banner",
        "ban_text_check_result",
        "can_be_deleted",
        "incompatible",
        "can_subscribe",
        "subscriptions",
        "time_subscribed",
        "favourited",
        "followers",
        "lifetime_subscriptions",
        "lifetime_favourited",
        "lifetime_followers",
        "lifetime_playtime",
        "lifetime_playtime_sessions",
        "image_width",
        "image_height",
        "image_url",
        "shortcut_id",
        "shortcut_name",
        "consumer_shortcut_id",
        "tags",
        "kv_tags",
        "reports",
        "upvotes",
        "downvotes",
        "score",
        "playtime_stats",
        "for_sale_data",
        "metadata",
        "language",
        "maybe_inappropriate_sex",
        "maybe_inappropriate_violence",
        "reactions",
    )

    # TODO needs
    # Client.create_published_file() -> PublishedFile
    # although this needs steammessages_cloud.proto to be compiled and an api around that

    def __init__(self, state: ConnectionState, proto: PublishedFileDetails, author: Authors):
        self._state = state
        self.id = proto.publishedfileid
        """The file's id."""
        self.name = proto.title
        """The file's name."""
        self.author = author
        """The file's author."""
        self.game = StatefulGame(state, id=proto.consumer_appid, name=proto.app_name)
        """The file's game."""
        self.created_with = StatefulGame(state, id=proto.creator_appid)
        """The file's created_with."""
        self.created_at = DateTime.from_timestamp(proto.time_created)
        """The time the file was created at."""
        self.updated_at = DateTime.from_timestamp(proto.time_updated)
        """The time the file was last updated at."""
        self.url = str(
            URL.COMMUNITY / "sharedfiles/filedetails" % {"id": self.id}
        )  # proto.url is seemingly always empty
        """The file's url."""
        self.manifest_id = proto.hcontent_file
        """The file's manifest id."""
        self.revision = PublishedFileRevision.try_value(proto.revision)
        """The file's revision."""
        self.available_revisions = [PublishedFileRevision.try_value(revision) for revision in proto.available_revisions]
        """The file's available_revisions."""
        self.change_number = proto.revision_change_number
        """This revision's change number."""
        self.type = PublishedFileType.try_value(proto.file_type)
        """The file's type."""
        self.views = proto.views
        """The file's views."""
        self.youtube_id = proto.youtubevideoid
        """The file's youtube_id."""
        self.description = proto.file_description
        """The file's description."""
        self.visibility = PublishedFileVisibility.try_value(proto.visibility)
        """The file's visibility."""
        self.flags = proto.flags  # TODO figure out what these are
        """The file's flags."""
        self.accepted = proto.workshop_accepted
        """Whether the file has been accepted."""
        self.show_subscribe_all = proto.show_subscribe_all
        """The file's show_subscribe_all."""
        self.number_of_developer_comments = proto.num_comments_developer
        """The file's number of comments from the developer."""
        self.number_of_public_comments = proto.num_comments_public
        """The file's number of public comments."""
        self.spoiler = proto.spoiler_tag
        """Whether the file is marked as a spoiler."""
        self.children = [
            PublishedFileChild(state, p.publishedfileid, p.sortorder, PublishedFileType.try_value(p.file_type), self)
            for p in proto.children
        ]
        """The file's children."""

        self.filename = proto.filename
        """The file's filename."""
        self.size = proto.file_size
        """The file's size."""
        self.cdn_url = proto.file_url
        """The file's cdn_url."""

        self.preview = PreviewInfo(state, proto.hcontent_preview, proto.preview_file_size, proto.preview_url)
        """The file's preview."""
        self.previews = [
            DetailsPreview(state, p.previewid, p.sortorder, p.url, p.size, p.filename, p.preview_type)
            for p in proto.previews
        ]
        """All the file's previews."""
        self.banned = proto.banned
        """Whether the file is banned."""
        self.ban_reason = proto.ban_reason if self.banned else None
        """The file's ban_reason."""
        self.banner = proto.banner if self.banned else None
        """The file's banner."""
        self.ban_text_check_result = proto.ban_text_check_result
        """The file's ban_text_check_result."""

        self.can_be_deleted = proto.can_be_deleted
        """The file's can_be_deleted."""
        self.incompatible = proto.incompatible
        """The file's incompatible."""

        self.can_subscribe = proto.can_subscribe
        """The file's can_subscribe."""
        self.subscriptions = proto.subscriptions
        """The file's subscriptions."""
        self.time_subscribed = proto.time_subscribed
        """The file's time_subscribed."""

        self.favourited = proto.favorited
        """The current number of people that have favourited the file."""
        self.followers = proto.followers
        """The current number of followers the file has."""
        self.lifetime_subscriptions = proto.lifetime_subscriptions
        """The file's lifetime subscriptions."""
        self.lifetime_favourited = proto.lifetime_favorited
        """The total number of people that have favourited the file."""
        self.lifetime_followers = proto.lifetime_followers
        """The total number of followers the file has."""
        self.lifetime_playtime = proto.lifetime_playtime
        """The file's lifetime playtime."""
        self.lifetime_playtime_sessions = proto.lifetime_playtime_sessions
        """The file's lifetime playtime sessions."""

        self.image_width = proto.image_width
        """The file's image's width."""
        self.image_height = proto.image_height
        """The file's image's height."""
        self.image_url = proto.image_url
        """The file's image's url."""

        self.shortcut_id = proto.shortcutid
        """The file's shortcut ID."""
        self.shortcut_name = proto.shortcutname
        """The file's shortcut name."""
        self.consumer_shortcut_id = proto.consumer_shortcutid
        """The file's consumer's shortcut ID."""

        self.tags = proto.tags
        """The file's tags."""
        self.kv_tags = {tag.key: tag.value for tag in proto.kvtags}
        """The file's key-value tags"""

        self.reports = proto.num_reports
        """The number of reports the file has."""
        self.upvotes = proto.vote_data.votes_up
        """The number of upvotes the file has."""
        self.downvotes = proto.vote_data.votes_down
        """The number of downvotes the file has."""
        self.score = proto.vote_data.score
        """The file's score."""
        self.playtime_stats = proto.playtime_stats
        """The file's playtime stats."""
        self.for_sale_data = proto.for_sale_data
        """The file's sale data."""

        self.metadata = proto.metadata
        """Any metadata attached to the file."""
        self.language = Language.try_value(proto.language)
        """The language the file is in."""
        self.maybe_inappropriate_sex = proto.maybe_inappropriate_sex
        """Whether the file may contain sexually inappropriate content."""
        self.maybe_inappropriate_violence = proto.maybe_inappropriate_violence
        """Whether the file may contain violent content which may be inappropriate for some."""

        self.reactions = [AwardReaction(state, reaction) for reaction in proto.reactions]
        """The file's reactions."""

    def __repr__(self) -> str:
        attrs = ("name", "id", "author", "game", "manifest_id", "change_number")
        resolved = [f"{name}={getattr(self, name)!r}" for name in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    @property
    def _commentable_kwargs(self) -> _CommentableKwargs:
        return {
            "id64": self.author.id64,
            "thread_type": 9,
        }

    async def manifest(self) -> Manifest:
        """The manifest associated with this published file."""
        return await self._state.fetch_manifest(self.game.id, self.manifest_id, self.game.id, self.name)

    @asynccontextmanager
    async def open(self) -> AsyncGenerator[BytesIO, None]:
        if self.type not in (PublishedFileType.Art, PublishedFileType.SteamVideo):
            raise NotImplementedError(f"Cannot open {self.type}")

        async with self._state.http._session.get(self.cdn_url) as r:
            io = BytesIO(await r.read())
            io.name = self.filename
            yield io

    async def fetch_children(
        self, *, revision: PublishedFileRevision = PublishedFileRevision.Default, language: Language | None = None
    ) -> list[PublishedFile]:
        """Fetches this published file's children.

        Parameters
        ----------
        revision
            The revision to fetch.
        language
            The language to fetch children in. If ``None``, the current language is used.
        """
        results = await self._state.fetch_published_files((c.id for c in self.children), revision, language)
        # assert all(results)   # this is safe, hopefully this can be added if HKT happens
        return results  # type: ignore

    async def parents(
        self,
        *,
        revision: PublishedFileRevision = PublishedFileRevision.Default,
        language: Language | None = None,
    ) -> list[PublishedFile]:
        """Fetches this published file's parents.

        Parameters
        ----------
        revision
            The revision to fetch.
        language
            The language to fetch parents in. If ``None``, the current language is used.
        """
        cursor = "*"
        more = True
        parents: list[PublishedFile] = []
        authors: set[int] = set()

        while more:
            proto = await self._state.fetch_published_file_parents(self.id, revision, language, cursor)
            more = len(parents) < proto.total

            for file in proto.publishedfiledetails:
                author: Authors = file.creator  # type: ignore
                parents.append(PublishedFile(self._state, file, author))
                authors.add(file.creator)

        for author in await self._state._maybe_users(authors):
            for parent in parents:
                if parent.author == author.id64:
                    parent.author = author

        return parents

    async def upvote(self) -> None:
        """Upvotes this published file."""
        await self._state.upvote_published_file(self.id, True)

    async def downvote(self) -> None:
        """Downvotes this published file."""
        await self._state.upvote_published_file(self.id, False)

    async def subscribe(self) -> None:
        """Subscribes to this published file's changes."""
        await self._state.subscribe_to_published_file(self.id)

    async def unsubscribe(self) -> None:
        """Unsubscribes from this published file's changes."""
        await self._state.unsubscribe_from_published_file(self.id)

    async def is_subscribed(self) -> bool:
        """Whether the client user is subscribed to this published file."""
        return await self._state.is_subscribed_to_published_file(self.id)

    async def add_child(self, child: PublishedFile) -> None:
        """Adds a child to this published file.

        Parameters
        ----------
        child
            The child to add.
        """
        await self._state.add_published_file_child(self.id, child.id)

    async def remove_child(self, child: PublishedFile) -> None:
        """Removes a child from this published file.

        Parameters
        ----------
        child
            The child to remove.
        """
        await self._state.remove_published_file_child(self.id, child.id)

    async def history(self, *, language: Language | None = None) -> list[PublishedFileChange]:
        """Fetches this published file's history.

        Parameters
        ----------
        language
            The language to fetch history in. If ``None``, the current language is used.
        """
        changes = await self._state.fetch_published_file_history(self.id, language)
        return [
            PublishedFileChange(change.change_description, DateTime.from_timestamp(change.timestamp))
            for change in changes
        ]

    async def fetch_history_entry(self, at: datetime, *, language: Language | None = None) -> PublishedFileChange:
        """Fetches the history entry at a given time.

        Parameters
        ----------
        at
            The time to fetch the history entry at.
        language
            The language to fetch history entries in. If ``None``, the current language is used.
        """
        change = await self._state.fetch_published_file_history_entry(self.id, at, language)
        return PublishedFileChange(change.change_description, at)

    # async def relationship(self):
    #     ...  # literally no clue what this does

    # async def friends_who_favourited(self):
    #     "https://steamcommunity.com/sharedfiles/friendswhofavoritedfile?id=2120834324&appid=440"
    #     # requires avatar hash property might wait till v1

    async def edit(
        self,
        *,
        name: str | None = None,
        description: str | None = None,
        visibility: PublishedFileVisibility | None = None,
        tags: Sequence[str] = (),
        filename: str | None = None,
        preview_filename: str | None = None,
    ) -> None:
        """Edits this published file.

        Parameters
        ----------
        name
            The new name of the file.
        description
            The new description of the file.
        visibility
            The new visibility of the file.
        tags
            The new tags of the file.
        filename
            The new filename of the file.
        preview_filename
            The new preview filename of the file.
        """
        try:
            preview_filename = self.previews[0].filename  # TODO test
        except IndexError:
            preview_filename = ""
        await self._state.edit_published_file(
            self.id,
            self.game.id,
            name if name is not None else self.name,
            description if description is not None else self.description,
            visibility or self.visibility,
            tags if tags is not None else [t.display_name for t in self.tags],
            filename if filename is not None else self.filename,
            preview_filename if preview_filename is not None else preview_filename,
        )
